from collections import defaultdict
import math
import os
import json
import uuid
import time
import socket
from pathlib import Path
from typing import Annotated
import re

import yaml

import arguably
from rich.console import Console

from kalavai_client.cluster import CLUSTER
from kalavai_client.api import run_api
from kalavai_client.env import (
    USER_COOKIE,
    USER_LOCAL_SERVER_FILE,
    CONTAINER_HOST_PATH,
    USER_COMPOSE_FILE,
    USER_HELM_APPS_FILE,
    USER_KUBECONFIG_FILE,
    USER_VPN_COMPOSE_FILE,
    USER_TEMPLATES_FOLDER,
    DOCKER_COMPOSE_GUI,
    USER_GUI_COMPOSE_FILE,
    user_path,
    resource_path,
)
from kalavai_client.core import (
    generate_worker_package,
    load_gpu_models,
    check_token,
    attach_to_pool,
    join_pool,
    create_pool,
    get_ip_addresses,
    pause_agent,
    resume_agent,
    stop_pool,
    TokenType,
    update_pool,
    get_user_spaces
)
from kalavai_client.utils import (
    check_gpu_drivers,
    load_template,
    run_cmd,
    user_confirm,
    generate_table,
    request_to_server,
    safe_remove,
    load_server_info,
    load_user_id,
    parse_key_value_pairs,
    request_to_api,
    store_server_info,
    has_api_details,
    CLUSTER_NAME_KEY,
    KALAVAI_AUTH,
    KALAVAI_API_URL_KEY,
    KALAVAI_API_KEY_KEY
)


LOCAL_TEMPLATES_DIR = os.getenv("LOCAL_TEMPLATES_DIR", None)
VERSION = 1
RESOURCE_EXCLUDE = ["ephemeral-storage", "hugepages-1Gi", "hugepages-2Mi", "pods"]
CORE_NAMESPACES = ["lws-system", "kube-system", "gpu-operator", "kalavai"]
RAY_LABEL = "kalavai.ray.name"
PVC_NAME_LABEL = "kalavai.storage.name"
VPN_COMPOSE_TEMPLATE = resource_path("kalavai_client/assets/vpn-template.yaml")
STORAGE_CLASS_NAME = "local-path"
STORAGE_ACCESS_MODE = ["ReadWriteOnce"]
DEFAULT_STORAGE_NAME = "pool-cache"
DEFAULT_STORAGE_SIZE = 20

    
console = Console()


######################
## HELPER FUNCTIONS ##
######################


def cleanup_local():
    console.log("Removing local cache files...")
    safe_remove(CONTAINER_HOST_PATH)
    safe_remove(USER_COMPOSE_FILE)
    safe_remove(USER_VPN_COMPOSE_FILE)
    safe_remove(USER_HELM_APPS_FILE)
    safe_remove(USER_KUBECONFIG_FILE)
    safe_remove(USER_LOCAL_SERVER_FILE)
    safe_remove(USER_TEMPLATES_FOLDER)
    #safe_remove(USER_GUI_COMPOSE_FILE)

def pre_join_check(node_name, server_url, server_key):
    # check with the server that we can connect
    try:
        nodes = request_to_server(
            force_url=server_url,
            force_key=server_key,
            method="get",
            endpoint="/v1/get_nodes",
            data={"node_names": [node_name]},
            server_creds=USER_LOCAL_SERVER_FILE,
            user_cookie=USER_COOKIE
        )
        return node_name not in nodes.keys()
    except Exception as e:
        console.log(f"[red]Error when connecting to kalavai service: {str(e)}")
        return False

def select_ip_address(subnet=None):
    ips = get_ip_addresses(subnet=subnet)
    if len(ips) == 1:
        return ips[0]

    while True:
        option = user_confirm(
            question="Select IP to advertise the node (needs to be visible to other nodes)",
            options=ips
        )
        if option is not None:
            break
        else:
            console.log("[red] Input error")
    return ips[option]

def select_gpus(message):
    console.log(f"[yellow]{message}")
    gpu_models = ["Any/None"]
    gpu_models_full = ["Any/None"]
    available_gpus = load_gpu_models()
    for _, gpus in available_gpus:
        for gpu in gpus["gpus"]:
            #status = "free" if "ready" in gpu else "busy"
            memory = math.floor(int(gpu['memory'])/1000)
            gpu_models.append(gpu['model'])
            gpu_models_full.append(f"{gpu['model']} ({memory}GB) (in use: {gpus['available'] == 0})" )
    
    while True:
        options = user_confirm(
            question=" ",
            options=gpu_models_full,
            multiple=True
        )
        if options is not None:
            if 0 in options:
                ids = None
            else:
                ids = ",".join([gpu_models[i] for i in options])
            break
    return ids

def select_token_type():
    options = ["Admin", "User (deploy jobs)", "Worker (read only)"]
    
    while True:
        choice = user_confirm(
            question="What type of access are you granting?",
            options=options,
            multiple=False
        )
        if choice is not None:
            break
    return {"admin": choice == 0, "user": choice == 1, "worker": choice == 2}

def input_gpus(non_interactive=False):
    num_gpus = 0
    try:
        has_gpus = check_gpu_drivers()
        if has_gpus:
            max_gpus = len(
                [r for r in run_cmd("nvidia-smi -L").decode().split("\n") if len(r.strip())>0]
            )
            if non_interactive:
                num_gpus = max_gpus
            else:
                num_gpus = user_confirm(
                    question=f"{max_gpus} NVIDIA GPU(s) detected. How many GPUs would you like to include?",
                    options=range(max_gpus+1)
                ) 
    except:
        console.log(f"[red]WARNING: error when fetching NVIDIA GPU info. GPUs will not be used on this local machine")
    return num_gpus

def show_connection_suggestion():
    console.log("[red]Not connected to a local or remote pool")
    console.log("Suggestions:")
    console.log("- Create a pool: [yellow]kalavai pool start")
    console.log("- Join a pool: [yellow]kalavai pool join")
    console.log("- Connect to a remote pool: [yellow]kalavai pool connect")

##################
## CLI COMMANDS ##
##################

@arguably.command
def pool__spaces(*others):
    """List available user spaces in the pool"""
    spaces = request_to_api(
        method="GET",
        endpoint="/get_available_user_spaces"
    )
    console.log("Available user spaces")

    if "error" in spaces:
        console.log(f"[red]Error: {spaces}")
    for space in spaces:
        console.log(f"-> {space}")

@arguably.command
def gui__start(
    *others,
    version="latest"
):
    """Run GUI (docker) and kalavai core backend (api)"""
    ports_needed = 2
    # find 2 available ports
    ip = socket.gethostbyname (socket.gethostname())
    ports = []
    for port in range(49153,65535):
        try:
            serv = socket.socket(socket.AF_INET,socket.SOCK_STREAM) # create a new socket
            serv.bind((ip, port)) # bind socket with address
            serv.close()
            ports.append(port)
        except:
            #port closed
            pass
        if len(ports) >= ports_needed:
            break
    
    if len(ports) < ports_needed:
        # if not found, error
        console.log(f"[red]Cannot initialise GUI: Could not find {ports_needed} free ports in your machine")    
        return
    console.log(f"Using ports: {ports}")
    kalavai_api_url = load_server_info(data_key=KALAVAI_API_URL_KEY, file=USER_LOCAL_SERVER_FILE)
    if kalavai_api_url is None:
        show_connection_suggestion()
        return

    user_key = load_server_info(data_key=KALAVAI_API_KEY_KEY, file=USER_LOCAL_SERVER_FILE)
    if user_key is not None:
        console.log(f"[green]Using user key: {user_key}")
    
    values = {
        "gui_frontend_port": ports[0],
        "gui_backend_port": ports[1],
        "kalavai_api_url": kalavai_api_url,
        "path": user_path("", create_path=True),
        "protected_access": user_key,
        "gui_version": version
    }
    compose_yaml = load_template(
        template_path=DOCKER_COMPOSE_GUI,
        values=values)
    with open(USER_GUI_COMPOSE_FILE, "w") as f:
        f.write(compose_yaml)
    
    run_cmd(f"docker compose --file {USER_GUI_COMPOSE_FILE} up -d")

    console.log(f"[green]Loading GUI, may take a few minutes. It will be available at http://localhost:{ports[0]}")
    console.log("Run [yellow]kalavai gui stop[white] to stop running the GUI")

@arguably.command
def gui__stop():
    try:
        run_cmd(f"docker compose --file {USER_GUI_COMPOSE_FILE} down")
    except Exception as e:
        console.log(f"Error when stopping GUI: {str(e)}. IGNORE if GUI was not running before")
    
@arguably.command
def auth(user_key, *others):
    """
    [AUTH] (For public clusters only) Log in to Kalavai server.
    """
    KALAVAI_AUTH.save_auth(user_key)
    if KALAVAI_AUTH.is_authenticated():
        console.log(f"[green]User key stored")
    else:
        console.log(f"[red]Invalid user key")

@arguably.command
def logout(*others):
    """
    Log out of Kalavai server.
    """
    KALAVAI_AUTH.clear_auth()
    console.log(f"[green]User key removed")

@arguably.command
def pool__package_worker(output_file, *others, platform="amd64", num_gpus=0, ip_address="0.0.0.0", node_name=None, storage_compatible=True):
    """
    [AUTH]Package a worker for distribution (docker compose only)
    """

    if not CLUSTER.is_seed_node():
        console.log(f"[red]You can only create workers from a seed node")
        return
    
    compose = generate_worker_package(
        target_platform=platform,
        num_gpus=num_gpus,
        ip_address=ip_address,
        node_name=node_name,
        storage_compatible=storage_compatible
    )

    if "error" in compose:
        console.log(f"[red]{compose['error']}")
    else:
        console.log(f"[green]Worker package created: {output_file}")
        with open(output_file, "w") as f:
            f.write(compose)

@arguably.command
def pool__credentials(*others):
    """
    Show credentials details for a remote machine to connect to local pool
    """
    if not has_api_details():
        show_connection_suggestion()
        return
    
    data = request_to_api(
        method="GET",
        endpoint="/get_pool_credentials"
    )
    try:
        url = data[KALAVAI_API_URL_KEY]
        key = data[KALAVAI_API_KEY_KEY]

        console.log(f"[green]Kalavai API URL: {url}")
        console.log(f"[green]Kalavai API Key: {key}")
        console.log("\n")
        console.log(f"Run the following from a remote machine to connect to this pool")
        console.log(f"[yellow]kalavai pool connect {url} {key}")
    except Exception as e:
        console.log(f"[red]Error: {e}")

@arguably.command
def pool__connect(url: str, key: str, *others):
    """
    Connect local instance to a remote kalavai pool.

    This does not join the physical machine to the pool, just connects to 
    the pool API.
    """
    url_with_port_regex = re.compile(
        r'^https?:\/\/'         # Start with 'http://' or 'https://'
        r'(([a-zA-Z0-9.-]+)|(localhost))' # **CAPTURE GROUP 1: The address/domain/localhost**
        r':\d{1,5}$'            # Colon and port
    )

    match = url_with_port_regex.match(url)
    if not match:
        console.log("[red]Unexpected url format, expected http://<address>:<port>")
        return
    ip_address = match.group(1)
    
    # - check if user connected already --> reject if so
    if has_api_details():
        option = user_confirm(
            question="You seem to be connected to an instance already. Are you sure you want to join a new one? You will lose access to the current pool and this cannot be reversed",
            options=["no", "yes"]
        )
        if option == 0:
            console.log("[green]Nothing happened.")
            return
        else:
            console.log("[yellow]Overwriting old config...")
    
    # store new details for remote API
    store_server_info(
        file=USER_LOCAL_SERVER_FILE,
        kalavai_api_url=url,
        kalavai_api_key=key,
        server_ip=ip_address,
        auth_key=None,
        watcher_service=None,
        node_name=None,
        cluster_name=None
    )
    console.log(f"[green]Local instance connected to {url}")

@arguably.command
def pool__start(
    *others, 
    pool_config_file=None,
    apps: list=None,
    mtu: str="",
    watcher_image_tag: str=None,
    platform="amd64",
    ip_address: str=None,
    lb_address: str=None,
    location: str=None,
    non_interactive: bool=False,
    kalavai_api_version: str=None,
    node_labels: Annotated[dict, arguably.arg.handler(parse_key_value_pairs)] = {}
):

    """
    Start Kalavai pool and start/resume sharing resources.

    Args:
        *others: all the other positional arguments go here
    """

    if CLUSTER.is_cluster_init():
        console.log(f"[white] You are already connected to {load_server_info(data_key=CLUSTER_NAME_KEY, file=USER_LOCAL_SERVER_FILE)}. Enter [yellow]kalavai pool stop[white] to exit and join another one.")
        return
    
    if non_interactive and all([value is None for value in [location, lb_address, ip_address]]):
        console.log("[red]In --non-interactive mode without --location, one of --lb-address or --ip-address must be set")
        return

    if node_labels:
        console.log(f"[blue]Configuration received: {node_labels}")
    
    # User acknowledgement
    if not non_interactive:
        option = user_confirm(
            question="Kalavai will now create a pool and a local worker using docker. This won't modify your system. Are you happy to proceed?",
            options=["no", "yes"]
        )
        if option == 0:
            console.log("Installation was cancelled and did not complete.")
            return
    
    # select IP address (for external discovery)
    if ip_address is None and location is None:
        if non_interactive:
            ip_address = "0.0.0.0"
            console.log("[yellow]Using [green]0.0.0.0 [yellow]for server address")
        else:
            # local IP
            console.log(f"Scanning for valid IPs")
            ip_address = select_ip_address()
    
    console.log(f"Using {ip_address} address for server")

    console.log(f"[green]Creating pool, this may take a few minutes...")

    result = create_pool(
        ip_address=ip_address,
        lb_ip_address=lb_address,
        location=location,
        target_platform=platform,
        watcher_image_tag=watcher_image_tag,
        pool_config_file=pool_config_file,
        apps=apps,
        num_gpus=input_gpus(non_interactive=non_interactive),
        mtu=mtu,
        node_labels=node_labels,
        kalavai_api_version=kalavai_api_version
    )

    if "warning" in result:
        console.log(f"[yellow]Warning: {result['warning']}")

    if "error" in result:
        console.log(f"[red]{result}")

    return None

@arguably.command
def pool__token(*others, admin=False, user=False, worker=False):
    """
    Generate a join token for others to connect to your pool
    """
    # try:
    #     CLUSTER.validate_cluster()
    # except Exception as e:
    #     console.log(f"[red]Problems with your pool: {str(e)}")
    #     return
    if not has_api_details():
        show_connection_suggestion()
        return
    
    if not admin and not user and not worker:
        console.log(f"[red]Select at least one mode (--admin, --user or --worker)")
        return
    
    if admin:
        mode = TokenType.ADMIN
    elif user:
        mode = TokenType.USER
    else:
        mode = TokenType.WORKER

    #join_token = get_pool_token(mode=mode)
    join_token = request_to_api(
        method="GET",
        endpoint="/get_pool_token",
        params={"mode": mode.value}
    )

    if "error" in join_token:
        console.log(f"[red]{join_token}")
    else:
        console.log("[green]Join token:")
        print(join_token["token"])
    return join_token

@arguably.command
def pool__check_token(token, *others, public=False, verbose=False):
    """
    Utility to check the validity of a join token
    """
    
    result = check_token(token=token, public=public)
    if "error" in result:
        console.log(f"[red]Error in token: {result}")
        return False
    if verbose:
        console.log(json.dumps(result["data"], indent=2))
    
    console.log("[green]Token format is correct")
    return True

@arguably.command
def pool__join(
    token,
    *others,
    mtu="",
    platform="amd64",
    node_name=None,
    non_interactive=False,
    node_labels: Annotated[dict, arguably.arg.handler(parse_key_value_pairs)] = {},
    seed: bool=False,
    kalavai_api_version: str=None
):
    """
    Join Kalavai pool and start/resume sharing resources.

    Args:
        token: Pool join token
        *others: all the other positional arguments go here
        mtu: Maximum transmission unit
        platform: Target platform (default: amd64)
        node_name: Name for this node
        non_interactive: Run in non-interactive mode
        node_labels: Node labels as key=value pairs (e.g., "key1=value1,key2=value2")
        seed: if the node should join as an extra seed (for HA deployments)
    """
    
    # Process node labels if provided
    if node_labels:
        console.log(f"[blue]Configuration received: {node_labels}")
    
    # check that k3s is not running already in the host
    # k3s service running or preinstalled
    if CLUSTER.is_agent_running():
        console.log(f"[white] You are already connected to {load_server_info(data_key=CLUSTER_NAME_KEY, file=USER_LOCAL_SERVER_FILE)}. Enter [yellow]kalavai pool stop[white] to exit and join another one.")
        return
    
    # check that is not attached to another instance
    if not non_interactive:
        if os.path.exists(USER_LOCAL_SERVER_FILE):
            option = user_confirm(
                question="You seem to be connected to an instance already. Are you sure you want to join a new one?",
                options=["no", "yes"]
            )
            if option == 0:
                console.log("[green]Nothing happened.")
                return
    
    user_id = load_user_id()
    if user_id is None:
        console.log("You are not authenticated. If you want to authenticate your node, use [yellow]kalavai auth <user_key>")
    
    num_gpus = input_gpus(non_interactive=non_interactive)

    if not non_interactive:
        option = user_confirm(
            question="Docker compose ready. Would you like Kalavai to deploy it?",
            options=["no", "yes"]
        )
        if option == 0:
            console.log("[red]Installation aborted")
            return
    
    # select IP address (for external discovery)
    console.log(f"Scanning for valid IPs")
    if non_interactive:
        ip_address = "0.0.0.0"
        console.log("[yellow]Using [green]0.0.0.0 [yellow]for server address")
    else:
        ip_address = select_ip_address()
    
    console.log("Connecting worker to the pool...")
    result = join_pool(
        target_platform=platform,
        token=token,
        node_name=node_name,
        num_gpus=num_gpus,
        ip_address=ip_address,
        mtu=mtu,
        node_labels=node_labels,
        is_seed=seed,
        kalavai_api_version=kalavai_api_version
    )
    if "error" in result:
        console.log(f"[red]Error when connecting: {result}")
    else:
        console.log(f"[green] You are connected to {result}")

@arguably.command
def pool__stop(*others, skip_node_deletion=False):
    """
    Stop sharing your device and clean up. DO THIS ONLY IF YOU WANT TO REMOVE KALAVAI-CLIENT from your device.

    Args:
        *others: all the other positional arguments go here
    """
    console.log("[white] Stopping kalavai GUI...")
    gui__stop()
    console.log("[white] Stopping kalavai app...")
    result = stop_pool(skip_node_deletion=skip_node_deletion)
    if "error" in result:
        console.log(f"[red]{result['error']}")
    else:
        console.log(result)
        console.log("[white] Kalavai has stopped sharing your resources. Use [yellow]kalavai pool start[white] or [yellow]kalavai pool join[white] to start again!")

@arguably.command
def pool__pause(*others):
    """
    Pause sharing your device and make your device unavailable for kalavai scheduling.

    Args:
        *others: all the other positional arguments go here
    """
    # k3s stop locally
    if not CLUSTER.is_cluster_init():
        console.log("[red] Kalavai pool not running locally. Cannot pause, please run [yellow]kalavai pool start[red] to start a pool or [yellow]kalavai pool join[red] to join one first")
        return
    console.log("[white] Pausing kalavai app...")
    success = pause_agent()
    if "error" in success:
        console.log(f"[red] Error when stopping. {success['error']}")
    else:
        console.log("[white] Kalava sharing paused. Resume with [yellow]kalavai pool resume")
        

@arguably.command
def pool__resume(*others):
    """
    Resume sharing your device and make device available for kalavai scheduling.

    Args:
        *others: all the other positional arguments go here
    """
    if not CLUSTER.is_cluster_init():
        console.log("[red] Kalavai pool not running locally. Cannot resume, please run [yellow]kalavai pool start[red] to start a pool or [yellow]kalavai pool join[red] to join one first")
        return
    console.log("[white] Restarting sharing (may take a few minutes)...")
    success = resume_agent()
    if "error" in success:
        console.log(f"[red] Error when restarting. {success['error']}")
    else:
        console.log("[white] Kalava sharing resumed")

@arguably.command
def pool__gpus(*others, available=False):
    """
    Display GPU information from all connected nodes
    """

    if not has_api_details():
        show_connection_suggestion()
        return
    
    gpus = request_to_api(
        method="GET",
        endpoint="/fetch_gpus",
        params={"available": available}
    )
    if "error" in gpus:
        console.log(f"[red]Error when fetching gpus: {gpus}")
        return
    
    columns = ["Node", "Ready", "GPU(s)", "Available", "Total"]
    rows = []
    try:
        for gpu in gpus:
            rows.append([
                gpu["node"],
                str(gpu["ready"]),
                gpu["model"],
                str(gpu["available"]),
                str(gpu["total"])
            ])
        console.print(
            generate_table(columns=columns, rows=rows,end_sections=[n for n in range(len(rows))])
        )
    except Exception as e:
        console.log(f"[red]Error: {str(e)}")

@arguably.command
def pool__resources(*others):
    """
    Display information about resources on the pool
    """
    if not has_api_details():
        show_connection_suggestion()
        return
    
    data = request_to_api(
        method="GET",
        endpoint="/fetch_resources"
    )

    if "error" in data:
        console.log(f"[red]Error when connecting to kalavai service: {data}")
        return
    
    columns = []
    total_values = []
    available_values = []
    for col in data["total"].keys():
        if col in RESOURCE_EXCLUDE:
            continue
        columns.append(col)
        total_values.append(str(data["total"][col]))
        available_values.append(str(data["available"][col]))
    
    columns = [""] + columns
    total_values = ["Total"] + total_values
    available_values = ["Available"] + available_values
    
    rows = [
        tuple(available_values),
        tuple(total_values)
    ]
    console.print(
        generate_table(columns=columns, rows=rows, end_sections=[0, 1])
    )
        

@arguably.command
def pool__update(*others):
    """
    Update kalavai pool
    """
    result = update_pool(debug=True)

    if "error" in result:
        console.log(f"[red]{result['error']}")
    else:
        console.log(f"[green]{result}")

@arguably.command
def pool__logs(*others, tail: int=100):
    """
    Get the logs for the Kalavai API
    """
    if not has_api_details():
        show_connection_suggestion()
        return
    
    logs = request_to_api(
        method="GET",
        endpoint="/fetch_service_logs",
        params={"tail": tail}
    )

    for name, log in logs.items():
        console.log(f"[yellow]LOGS for service: {name}")
        for key, value in log.items():
            console.log(f"[yellow]{key}")
            console.log(json.dumps(value, indent=2))

@arguably.command
def pool__status(*others, log_file=None):
    """
    Run diagnostics on a local installation of kalavai
    * is pool installed
    * is agent running
    * is kube-watcher running
    * is lws running
    """
    logs = []

    logs.append("Getting deployment status...")

    if CLUSTER.is_seed_node():
        # seed node
        data = CLUSTER.diagnostics()
        logs.append(data)
    else:
        # worker node
        logs.append("Could not access node info. This info is only available to seed nodes. Ignore if you are on a worker node.")
    logs.append(f"Worker installed: {CLUSTER.is_cluster_init()}")

    logs.append(f"Worker running: {CLUSTER.is_agent_running()}")

    logs.append(f"Pool credentials present: {CLUSTER.validate_cluster()}")

    logs.append(f"Is connected to Kalavai API (local/external): {has_api_details()}")
    
    if log_file is not None:
        with open(log_file, "w") as f:
            for log in logs:
                f.write(log)
                f.write("\n")
        console.log(f"[green]Logs written to {log_file}")
    else:
        for log in logs:
            console.log(f"{log}\n")

@arguably.command
def pool__attach(token, *others, node_name=None):
    """
    Set creds in token on the local instance
    """

    if node_name is None:
        node_name = f"{socket.gethostname()}-{uuid.uuid4().hex[:6]}"
    
    # check that is not attached to another instance
    if os.path.exists(USER_LOCAL_SERVER_FILE):
        option = user_confirm(
            question="You seem to be connected to an instance already. Are you sure you want to join a new one?",
            options=["no", "yes"]
        )
        if option == 0:
            console.log("[green]Nothing happened.")
            return
    
    option = user_confirm(
        question="Docker compose ready. Would you like Kalavai to deploy it?",
        options=["no", "yes"]
    )
    if option == 0:
        console.log("Manually deploy the worker with the following command:\n")
        print(f"docker compose -f {USER_COMPOSE_FILE} up -d")
        return
    
    result = attach_to_pool(token=token, node_name=node_name)
    
    if "error" in result:
        console.log(f"[red]Error when attaching to pool: {result}")
        return
    # set status to schedulable
    console.log(f"[green] You are connected to {result}")


@arguably.command
def storage__create(name, storage, *others, force_namespace: str=None):
    """
    Create storage for the cluster
    """
    try:
        CLUSTER.validate_cluster()
    except Exception as e:
        console.log(f"[red]Problems with your pool: {str(e)}")
        return
    
    if force_namespace is not None:
        console.log("[WARNING][yellow]--force-namespace [white]requires an admin key. Request will fail if you are not an admin.")

    # Deploy PVC
    data = {
        "name": name,
        "labels": {
            PVC_NAME_LABEL: name,
            "kalavai.resource": "storage"
        },
        "access_modes": STORAGE_ACCESS_MODE,
        "storage_class_name": STORAGE_CLASS_NAME,
        "storage_size": storage
    }
    if force_namespace is not None:
        data["force_namespace"] = force_namespace

    try:
        result = request_to_server(
            method="post",
            endpoint="/v1/deploy_storage_claim",
            data=data,
            server_creds=USER_LOCAL_SERVER_FILE,
            user_cookie=USER_COOKIE
        )
        if "error" in result or "detail" in result:
            console.log(f"Error: {result}")
        else:
            console.log(f"Storage {name} ({storage}Gi) created")
    except Exception as e:
        console.log(f"[red]Error when connecting to kalavai service: {str(e)}")


@arguably.command
def storage__list(*other):
    """
    List existing storages deployed in the pool
    """
    try:
        CLUSTER.validate_cluster()
    except Exception as e:
        console.log(f"[red]Problems with your pool: {str(e)}")
        return

    try:
        result = request_to_server(
            method="post",
            endpoint="/v1/get_storage_usage",
            data={},
            server_creds=USER_LOCAL_SERVER_FILE,
            user_cookie=USER_COOKIE
        )

        columns = []
        rows = []
        for namespace, storages in result.items():
            for name, values in storages.items():
                columns = list(values.keys())
                rows.append([namespace, name] + [f"{v:.2f} MB" if "capacity" in k else str(v) for k, v in values.items()])
        
        if len(rows) == 0:
            console.log("[green] Storages have not been claimed yet (did you deploy any job using them?)")
            return
        columns = ["Owner", "Name"] + columns
        table = generate_table(columns=columns, rows=rows)
        console.log(table)

    except Exception as e:
        console.log(f"[red]Error when connecting to kalavai service: {str(e)}")
    
@arguably.command
def storage__delete(name, *others, force_namespace: str=None):
    """
    Delete storage by name
    """
    try:
        CLUSTER.validate_cluster()
    except Exception as e:
        console.log(f"[red]Problems with your pool: {str(e)}")
        return
    
    if force_namespace is not None:
        console.log("[WARNING][yellow]--force-namespace [white]requires an admin key. Request will fail if you are not an admin.")

    # deploy template with kube-watcher
    data = {
        "label": PVC_NAME_LABEL,
        "value": name
    }
    if force_namespace is not None:
        data["force_namespace"] = force_namespace
    try:
        result = request_to_server(
            method="post",
            endpoint="/v1/delete_labeled_resources",
            data=data,
            server_creds=USER_LOCAL_SERVER_FILE,
            user_cookie=USER_COOKIE
        )
        console.log(f"{result}")
    except Exception as e:
        console.log(f"[red]Error when connecting to kalavai service: {str(e)}")

@arguably.command
def node__list(*others):
    """
    Display information about nodes connected
    """
    if not has_api_details():
        show_connection_suggestion()
        return
    
    devices = request_to_api(
        method="GET",
        endpoint="/fetch_devices"
    )

    try:
        rows = []
        columns = ["Node name", "Memory Pressure", "Disk pressure", "PID pressure", "Ready", "Unschedulable"]
        for device in devices:
            rows.append([
                device["name"],
                str(device["memory_pressure"]),
                str(device["disk_pressure"]),
                str(device["pid_pressure"]),
                str(device["ready"]),
                str(device["unschedulable"])
            ])
        
        console.log("Nodes with 'unschedulable=True' will not receive workload")
        console.log("To make a node unschedulable (i.e. won't receive workloads) use [yellow]kalavai node cordon <node name>")
        console.log("To make a node schedulable (i.e. will receive workloads) use [yellow]kalavai node uncordon <node name>")
        console.print(
            generate_table(columns=columns, rows=rows)
        )
        
    except Exception as e:
        console.log(f"[red]Error when connecting to kalavai service: {str(e)}")


@arguably.command
def node__delete(node_name, *others):
    """
    Delete a node from the cluster
    """
    if not has_api_details():
        show_connection_suggestion()
        return
    
    result = request_to_api(
        method="POST",
        endpoint="/delete_nodes",
        json={"nodes": [node_name]}
    )
    
    #result = delete_nodes(nodes=[name])
    
    if "error" in result:
        console.log(f"[red]{result}")
    else:
        console.log(f"[green]{result}")

@arguably.command
def node__cordon(node_name, *others):
    """
    Cordon a particular node so no more work will be scheduled on it
    """
    if not has_api_details():
        show_connection_suggestion()
        return
    
    result = request_to_api(
        method="POST",
        endpoint="/cordon_nodes",
        params={"nodes": [node_name]}
    )
    if "error" in result:
        console.log(f"[red]{result['error']}")
    else:
        console.log(result)

@arguably.command
def node__uncordon(node_name, *others):
    """
    Uncordon a particular node to allow more work to be scheduled on it
    """
    if not has_api_details():
        show_connection_suggestion()
        return
    
    result = request_to_api(
        method="POST",
        endpoint="/uncordon_nodes",
        params={"nodes": [node_name]}
    )
    if "error" in result:
        console.log(f"[red]{result['error']}")
    else:
        console.log(result)

@arguably.command
def job__templates(*others):
    """
    Job templates integrated with kalavai. Use env var LOCAL_TEMPLATES_DIR to test local templates
    """
    if not has_api_details():
        show_connection_suggestion()
        return
    
    templates = request_to_api(
        method="GET",
        endpoint="/fetch_job_templates"
    )
    if "error" in templates:
        console.log(f"[red]Error when fetching templates: {templates}")
        return
    
    console.log("Templates available in the pool")
    console.log(templates)


@arguably.command
def job__run(template_name, *others, values: str=None, force_namespace: str=None):
    """
    Deploy and run a template job.

    Args:
        *others: all the other positional arguments go here
    """
    if not has_api_details():
        show_connection_suggestion()
        return
    
    if force_namespace is not None:
        console.log("[WARNING][yellow]--force-namespace [white]requires an admin key. Request will fail if you are not an admin.")

    if values is None:
        values_dict = {}
    else:
        if not Path(values).is_file():
            console.log(f"[red]Values file {values} was not found")

        with open(values, "r") as f:
            raw_values = yaml.load(f, Loader=yaml.SafeLoader)
            values_dict = {variable["name"]: variable['value'] for variable in raw_values}

    data = {
        "template_name": template_name,
        "values": values_dict
    }
    if force_namespace is not None:
        data["force_namespace"] = force_namespace
    results = request_to_api(
        method="POST",
        endpoint="/deploy_job",
        json=data
    )

    if "error" in results:
        console.log(f"[red]Error when deploying job: {str(results['error'])}")
    else:
        try:
            for job in results:
                success = len(job["result"]["successful"]) if "successful" in job["result"] else 0
                fail = len(job["result"]["failed"]) if "failed" in job["result"] else 0
                console.log(f"[green]Job ID {job['job_id']} deployed ({success} OK, {fail} failed)")
        except Exception as e:
            console.log(f"[red]Error: {str(e)}")
            console.log(results)

@arguably.command
def job__test(local_template_dir, *others, values, force_namespace: str=None):
    """
    Helper to test local templates, useful for development
    """
    if not has_api_details():
        show_connection_suggestion()
        return
    
    if not os.path.isfile(os.path.join(local_template_dir, "template.yaml")):
        console.log(f"[red]template.yaml not found under {local_template_dir}")
        return
    if not os.path.isfile(os.path.join(local_template_dir, "values.yaml")):
        console.log(f"[red]values.yaml not found under {local_template_dir}")
        return
    
    # load template
    with open(os.path.join(local_template_dir, "template.yaml"), "r") as f:
        template_str = f.read()
    with open(os.path.join(local_template_dir, "values.yaml"), "r") as f:
        defaults = f.read()
    
    # load values
    if not os.path.isfile(values):
        console.log(f"[red]--values ({values}) is not a valid local file")
        return
    with open(values, "r") as f:
        raw_values = yaml.load(f, Loader=yaml.SafeLoader)
        values_dict = {variable["name"]: variable['value'] for variable in raw_values}

    data = {
        "template_str": template_str,
        "values": values_dict,
        "default_values": defaults,
        "target_labels": {}
    }
    if force_namespace is not None:
        data["force_namespace"] = force_namespace
    
    results = request_to_api(
        method="POST",
        endpoint="/deploy_custom_job",
        json=data
    )
    
    if "error" in results:
        console.log(f"[red]Error when deploying job: {str(results['error'])}")
    else:
        try:
            for job in results:
                success = len(job["result"]["successful"]) if "successful" in job["result"] else 0
                fail = len(job["result"]["failed"]) if "failed" in job["result"] else 0
                console.log(f"[green]Job ID {job['job_id']} deployed ({success} OK, {fail} failed)")
        except Exception as e:
            console.log(f"[red]{results}")
            console.log(f"[red]Error: {str(e)}")
            

@arguably.command
def job__defaults(template_name, *others):
    """
    Fetch default values.yaml for a template job
    """
    if not has_api_details():
        show_connection_suggestion()
        return
    
    defaults = request_to_api(
        method="GET",
        endpoint="/fetch_job_defaults",
        params={"name": template_name}
    )
    metadata = request_to_api(
        method="GET",
        endpoint="/fetch_job_metadata",
        params={"name": template_name}
    )

    if "error" in defaults:
        console.log(f"[red]Error when fetching job defaults: {defaults}")
    print(
        json.dumps(defaults, indent=3)
    )
    print(
        "*****************",
        "Metadata",
        "*****************"
    )
    print(
        json.dumps(metadata, indent=3)
    )


@arguably.command
def job__delete(name, *others, force_namespace: str=None):
    """
    Delete job in the cluster
    """
    if not has_api_details():
        show_connection_suggestion()
        return
    
    if force_namespace is not None:
        console.log("[WARNING][yellow]--force-namespace [white]requires an admin key. Request will fail if you are not an admin.")
    
    data = {
        "name": name
    }
    if force_namespace is not None:
        data["force_namespace"] = force_namespace
    results = request_to_api(
        method="POST",
        endpoint="/delete_job",
        json=data
    )
    if "error" in results:
        console.log(f"[red]Error when deleting job: {results['error']}")
    else:
        console.log(f"{results}")


@arguably.command
def job__list(*others, force_namespace: str=None):
    """
    List jobs in the cluster
    """
    if not has_api_details():
        show_connection_suggestion()
        return
    data = {}
    if force_namespace is not None:
        data["force_namespace"] = force_namespace
    details = request_to_api(
        method="GET",
        endpoint="/fetch_job_details",
        params=data
    )

    if "error" in details:
        console.log(f"[red]{details}")
        return
    try:
        columns = ["Owner", "Deployment", "Workers", "Endpoint"]
        rows = [[job["owner"], job["name"], job["workers"], job["endpoint"]] for job in details]
        
        console.print(
            generate_table(columns=columns, rows=rows, end_sections=range(len(rows)))
        )
            
        console.log("Get logs with [yellow]kalavai job logs <name of deployment> [white](note it only works when the deployment is complete)")
    except Exception as e:
        console.log(f"[red]Error: {details}")
        console.log("[red]Error when querying backend")

@arguably.command
def job__logs(name, *others, pod_name=None, tail=100, force_namespace: str=None):
    """
    Get logs for a specific job
    """
    if not has_api_details():
        show_connection_suggestion()
        return
    
    if force_namespace is not None:
        console.log("[WARNING][yellow]--force-namespace [white]requires an admin key. Request will fail if you are not an admin.")

    data = {
        "job_name": name,
        "pod_name": pod_name,
        "tail": tail,
    }
    if force_namespace is not None:
        data["force_namespace"] = force_namespace
    results = request_to_api(
        method="GET",
        endpoint="/fetch_job_logs",
        params=data)

    if "error" in results:
        console.log(f"[red]{results}")
        return
    for label_match, data in results.items():
        for pod, info in data.items():
            if pod_name is not None and pod_name != pod:
                continue
            if "logs" not in info:
                console.log(f"[white] Logs not ready for {pod}")
            else:
                logs = info["logs"]
                describe = info["describe"]
                console.log(f"[yellow]Logs for {pod} in {describe['spec']['node_name']}")
                console.log(f"[green]{logs}")
            console.log("---------------------------")
            console.log("---------------------------")
            if "describe" not in info:
                console.log(f"[white] Description not ready for {pod}")
            else:
                console.log(f"[yellow]Status {pod} in {describe['spec']['node_name']}")
                console.log(f"[green]{describe}")

def app():
    user_path("", create_path=True)
    arguably.run()

if __name__ == "__main__":
    app()


from pydantic import BaseModel, Field
from typing import List, Dict, Union, Optional
from enum import Enum


class Job(BaseModel):
    owner: Optional[str] = "default"
    name: Optional[str] = None
    workers: Optional[str] = None
    endpoint: Optional[str] = None
    status: Optional[str] = None
    host_nodes: Optional[str] = None

class DeviceStatus(BaseModel):
    name: str
    memory_pressure: bool
    disk_pressure: bool
    pid_pressure: bool
    ready: bool
    unschedulable: bool

class GPU(BaseModel):
    node: str
    available: int
    total: int
    ready: bool
    model: str

class TokenType(str, Enum):
    ADMIN = "admin"
    USER = "user"
    WORKER = "worker"

class CreatePoolRequest(BaseModel):
    cluster_name: str = Field(description="Name of the cluster to create")
    ip_address: str = Field(description="IP address for the pool")
    num_gpus: int = Field(None, description="Number of GPUs to allocate")
    node_name: str = Field(None, description="Name of the node")
    location: str = Field(None, description="Geographic location of the pool")

class WorkerConfigRequest(BaseModel):
    node_name: str = Field(None, description="Name for the worker node")
    mode: int = Field(2, description="Access mode for the worker (admin, worker or user)")
    target_platform: str = Field("amd64", description="Target platform architecture for the worker (amd64 or arm64)")
    num_gpus: int = Field(0, description="Number of GPUs to use on the worker node")
    ip_address: str = Field("0.0.0.0", description="IP address of the worker node")
    storage_compatible: bool = Field(True, description="Whether to use the node's storage capacity for volumes")


class NodesActionRequest(BaseModel):
    nodes: list[str] = Field(None, description="List of node names to perform the action on, defaults to None")

class JoinPoolRequest(BaseModel):
    token: str = Field(description="Token to join the pool")
    ip_address: str = Field(None, description="IP address for the node")
    node_name: str = Field(None, description="Name of the node")
    num_gpus: int = Field(None, description="Number of GPUs to allocate")
    frontend: bool = Field(False, description="Whether this is a frontend request")

class JobDetailsRequest(BaseModel):
    jobs: list[Job] = Field(description="List of jobs to get details for")

class StopPoolRequest(BaseModel):
    skip_node_deletion: bool = Field(False, description="Whether to skip node deletion when stopping the pool")

class DeployJobRequest(BaseModel):
    template_name: str = Field(description="Name of the job template to use")
    values: dict = Field(description="Job configuration values")
    force_namespace: Optional[str] = Field(None, description="Optional namespace override")
    target_labels: Optional[dict[str, Union[str, List]]] = Field(None, description="Optional target node labels")

class CustomDeployJobRequest(BaseModel):
    template_str: str = Field(description="YAML str containing the custom template job to use")
    values: dict = Field(description="Job configuration values")
    default_values: str = Field(description="YAML str containing the default values for the template job to use")
    force_namespace: Optional[str] = Field(None, description="Optional namespace override")
    target_labels: Optional[dict[str, Union[str, List]]] = Field(None, description="Optional target node labels")

class DeleteJobRequest(BaseModel):
    name: str = Field(description="Name of the job to delete")
    force_namespace: Optional[str] = Field(None, description="Optional namespace override")

class NodeLabelsRequest(BaseModel):
    node_name: str = Field(description="Name of the node to add labels to")
    labels: Dict[str, str] = Field(description="Dictionary of labels to add to the node")



"""
Core kalavai service.
Used as a bridge between the kalavai-client app and the reflex frontend
"""
import os
from argparse import ArgumentParser
from fastapi import FastAPI, HTTPException, Depends, Query, Security
from fastapi.security.api_key import APIKeyHeader
from typing import Optional, List
from fastapi_mcp import FastApiMCP
from starlette.requests import Request
import uvicorn

from kalavai_client.core import Job
from kalavai_client.env import (
    KALAVAI_SERVICE_LABEL,
    KALAVAI_SERVICE_LABEL_VALUE
)
from kalavai_client.api_models import (
    CreatePoolRequest,
    JoinPoolRequest,
    StopPoolRequest,
    DeployJobRequest,
    DeleteJobRequest,
    JobDetailsRequest,
    CustomDeployJobRequest,
    NodesActionRequest,
    NodeLabelsRequest,
    WorkerConfigRequest
)
from kalavai_client.core import (
    create_pool,
    join_pool,
    attach_to_pool,
    stop_pool,
    fetch_devices,
    fetch_resources,
    fetch_job_names,
    fetch_gpus,
    fetch_job_details,
    fetch_job_logs,
    fetch_job_templates,
    fetch_job_defaults,
    fetch_pod_logs,
    deploy_job,
    deploy_test_job,
    delete_job,
    authenticate_user,
    load_user_session,
    user_logout,
    is_connected,
    is_agent_running,
    is_server,
    pause_agent,
    resume_agent,
    get_ip_addresses,
    get_pool_token,
    delete_nodes,
    cordon_nodes,
    uncordon_nodes,
    add_node_labels,
    get_node_labels,
    generate_worker_package,
    get_user_spaces,
    get_space_quota,
    get_pool_credentials,
    TokenType
)

app = FastAPI(
    title="Kalavai Bridge API",
    description="API for managing Kalavai pools, jobs, and nodes",
    version="1.0.0",
    docs_url="/docs",
    redoc_url="/redoc",
)
API_KEY_NAME = "X-API-Key"
api_key_header = APIKeyHeader(name=API_KEY_NAME, auto_error=False)
MASTER_API_KEY = os.getenv("MASTER_API_KEY", None)
FORCED_USER_SPACE_NAME = os.getenv("FORCED_USER_SPACE_NAME", None)


################################
## API Key Validation methods ##
################################
async def verify_api_key(api_key_header: str = Security(api_key_header)):
    """
    Verify the API key against the master key.
    
    Args:
        api_key_header: The API key from the request header
        
    Returns:
        The API key if valid
        
    Raises:
        HTTPException: If the API key is invalid or missing
    """
    if MASTER_API_KEY is None:
        return api_key_header
    
    if not api_key_header:
        raise HTTPException(
            status_code=401,
            detail="API Key is missing",
            headers={"WWW-Authenticate": "ApiKey"},
        )
    
    if api_key_header != MASTER_API_KEY:
        raise HTTPException(
            status_code=403,
            detail="Invalid API Key",
            headers={"WWW-Authenticate": "ApiKey"},
        )
    
    return api_key_header 

@app.post("/create_pool", 
    operation_id="create_pool",
    summary="Create a new Kalavai compute pool",
    tags=["pool_management"],
    description="Creates a new distributed compute pool that allows multiple nodes to join and share GPU resources. The pool acts as a Kubernetes cluster where users can deploy and manage machine learning jobs across multiple devices.",
    response_description="Result of pool creation")
def pool_create(request: CreatePoolRequest, api_key: str = Depends(verify_api_key)):
    """
    Create a new pool with the following parameters:
    
    - **cluster_name**: Name of the cluster
    - **ip_address**: IP address for the pool
    - **num_gpus**: Number of GPUs to allocate
    - **node_name**: Name of the node
    - **location**: Location of the pool
    """
    result = create_pool(
        cluster_name=request.cluster_name,
        ip_address=request.ip_address,
        num_gpus=request.num_gpus,
        node_name=request.node_name,
        location=request.location,
    )
    return result

@app.post("/join_pool",
    operation_id="join_pool",
    summary="Join an existing Kalavai pool as a compute node",
    description="Joins a running Kalavai pool by providing a valid join token. This endpoint registers the current machine as a compute node in the pool, making its GPU resources available for job scheduling. The node will receive workloads based on the pool's scheduling policy.",
    tags=["pool_management"],
    response_description="Result of joining the pool")
def pool_join(request: JoinPoolRequest, api_key: str = Depends(verify_api_key)):
    """
    Join a pool with the following parameters:
    
    - **token**: Pool join token
    - **ip_address**: IP address for the node
    - **node_name**: Name of the node
    - **num_gpus**: Number of GPUs to allocate
    """
    result = join_pool(
        token=request.token,
        num_gpus=request.num_gpus,
        node_name=request.node_name,
        ip_address=request.ip_address
    )
    return result

@app.post("/attach_to_pool",
    operation_id="attach_to_pool",
    summary="Attach to a pool for management purposes",
    description="Attaches to an existing Kalavai pool for administrative and monitoring purposes without contributing compute resources. This is typically used by frontend applications or management tools that need to interact with the pool but don't provide GPU resources.",
    tags=["pool_management"],
    response_description="Result of attaching to the pool")
def pool_attach(request: JoinPoolRequest, api_key: str = Depends(verify_api_key)):
    """
    Attach to a pool with the following parameters:
    
    - **token**: Pool token
    - **node_name**: Name of the node
    - **frontend**: Whether this is a frontend request
    """
    result = attach_to_pool(
        token=request.token,
        node_name=request.node_name,
        frontend=request.frontend
    )
    return result

@app.post("/stop_pool",
    operation_id="stop_pool",
    summary="Stop and clean up the current Kalavai pool",
    description="Gracefully shuts down the current Kalavai pool, terminating all running jobs and optionally removing all compute nodes from the cluster. This operation is irreversible and will disconnect all nodes from the pool.",
    tags=["pool_management"],
    response_description="Result of stopping the pool")
def pool_stop(request: StopPoolRequest, api_key: str = Depends(verify_api_key)):
    """
    Stop the pool with the following parameters:
    
    - **skip_node_deletion**: Whether to skip node deletion
    """
    result = stop_pool(
        skip_node_deletion=request.skip_node_deletion
    )
    return result

@app.post("/delete_nodes",
    operation_id="delete_nodes",
    summary="Remove specific nodes from the pool",
    description="Removes specified compute nodes from the Kalavai pool. This operation will terminate any jobs running on the target nodes and clean up their resources. Use with caution as it may interrupt running workloads.",
    tags=["pool_management"],
    response_description="Result of node deletion")
def device_delete(request: NodesActionRequest, api_key: str = Depends(verify_api_key)):
    """
    Delete nodes with the following parameters:
    
    - **nodes**: List of node names to delete
    """
    result = delete_nodes(
        nodes=request.nodes
    )
    return result

@app.post("/cordon_nodes",
    operation_id="cordon_nodes",
    summary="Mark nodes as unschedulable",
    description="Marks specified nodes as unschedulable, preventing new jobs from being assigned to them while allowing existing jobs to complete. This is useful for maintenance operations or when you want to gradually remove nodes from the pool.",
    tags=["pool_management"],
    response_description="Result of cordoning nodes")
def device_cordon(request: NodesActionRequest, api_key: str = Depends(verify_api_key)):
    """
    Cordon nodes with the following parameters:
    
    - **nodes**: List of node names to cordon
    """
    result = cordon_nodes(
        nodes=request.nodes
    )
    return result

@app.post("/uncordon_nodes",
    operation_id="uncordon_nodes",
    summary="Mark nodes as schedulable again",
    description="Re-enables job scheduling on previously cordoned nodes, allowing them to receive new workloads. This reverses the effect of the cordon operation.",
    tags=["pool_management"],
    response_description="Result of uncordoning nodes")
def device_uncordon(request: NodesActionRequest, api_key: str = Depends(verify_api_key)):
    """
    Uncordon nodes with the following parameters:
    
    - **nodes**: List of node names to uncordon
    """
    result = uncordon_nodes(
        nodes=request.nodes
    )
    return result

@app.get("/get_pool_token",
    operation_id="get_pool_token",
    summary="Generate a token for pool access",
    description="Generates a secure token that can be used to join to the current Kalavai pool. Different token types provide different levels of access - join tokens allow nodes to contribute resources, while attach tokens allow management access.",
    tags=["auth"],
    response_description="Pool token")
def get_token(mode: TokenType, api_key: str = Depends(verify_api_key)):
    """
    Get pool token with the following parameters:
    
    - **mode**: Token type mode
    """
    return get_pool_token(mode=TokenType(mode))

@app.get("/get_pool_credentials",
    operation_id="get_pool_credentials",
    summary="Generate credentials for remote pool access",
    description="Generates credentials that can be used to connect to the current Kalavai pool.",
    tags=["auth"],
    response_description="Pool credentials")
def get_credentials(api_key: str = Depends(verify_api_key)):
    """
    Get pool credentials for remote connectivity
    """
    return get_pool_credentials()

@app.post("/generate_worker_config",
    operation_id="generate_worker_config",
    summary="Generate a config file for a remote worker to connect to the pool",
    description="Generate a config file for a remote worker to connect to the pool. Different token types provide different levels of access - join tokens allow nodes to contribute resources, while attach tokens allow management access.",
    tags=["pool_management"],
    response_description="Worker config file")
def generate_worker_config(request: WorkerConfigRequest, api_key: str = Depends(verify_api_key)):
    return generate_worker_package(
        node_name=request.node_name,
        mode=TokenType(request.mode),
        target_platform=request.target_platform,
        num_gpus=request.num_gpus,
        ip_address=request.ip_address,
        storage_compatible=request.storage_compatible)

@app.get("/fetch_devices",
    operation_id="fetch_devices",
    summary="Get list of all compute devices in the pool",
    description="Retrieves information about all compute devices (nodes) currently connected to the Kalavai pool, including their status, available resources, and current workload distribution.",
    tags=["info"],
    response_description="List of devices")
def get_devices(api_key: str = Depends(verify_api_key)):
    """Get list of available devices"""
    return fetch_devices()

@app.get("/fetch_service_logs",
    operation_id="fetch_service_logs",
    summary="Get logs for the kalavai API service",
    description="Get logs for the kalavai API service, including internal logs, debugging messages and status of the service.",
    tags=["info"],
    response_description="Logs")
def get_service_logs(tail: int=100, api_key: str = Depends(verify_api_key)):
    return fetch_pod_logs(label_key=KALAVAI_SERVICE_LABEL, label_value=KALAVAI_SERVICE_LABEL_VALUE, force_namespace="kalavai", tail=tail)

@app.get("/fetch_resources",
    operation_id="fetch_resources",
    summary="Get resource utilization for specific nodes",
    description="Retrieves detailed resource information (CPU, memory, GPU usage) for the pool; optionally for a list of specified nodes in the pool (as {'nodes': node_list}). This helps monitor resource utilization and plan workload distribution.",
    tags=["info"],
    response_description="Resource information")
def resources(request: Optional[NodesActionRequest]=NodesActionRequest(), api_key: str = Depends(verify_api_key)):
    """Get available resources"""
    return fetch_resources(node_names=request.nodes)

@app.get("/fetch_job_names",
    operation_id="fetch_job_names",
    summary="Get list of all jobs (model deployments) in the pool",
    description="Retrieves the names of all jobs and models currently deployed or scheduled in the Kalavai pool. This provides an overview of all workloads in the system.",
    tags=["info"],
    response_description="List of job names")
def job_names(api_key: str = Depends(verify_api_key)):
    """Get list of job names"""
    return fetch_job_names()

@app.get("/fetch_gpus",
    operation_id="fetch_gpus",
    summary="Get GPU information across the pool",
    description="Retrieves detailed information about all GPUs in the Kalavai pool, including their availability status, current utilization, and which jobs are using them. Can filter to show only available GPUs.",
    tags=["info"],
    response_description="List of GPUs")
def gpus(available: bool = False, api_key: str = Depends(verify_api_key)):
    """
    Get list of GPUs with the following parameters:
    
    - **available**: Whether to show only available GPUs
    """
    return fetch_gpus(available=available)

@app.get("/fetch_job_details",
    operation_id="fetch_job_details",
    summary="Get detailed information about specific job and model deployments",
    description="Retrieves comprehensive information about jobs or models including their status, resource usage, runtime, and configuration. Useful for monitoring and debugging job execution.",
    tags=["info"],
    response_description="Job details")
def job_details(force_namespace: str = Query(None), api_key: str = Depends(verify_api_key)):
    """Get job details"""
    return fetch_job_details(force_namespace=force_namespace)

@app.get("/fetch_job_logs",
    operation_id="fetch_job_logs",
    summary="Get execution logs for a specific job",
    description="Retrieves the execution logs for a specified job, providing real-time or historical output from the job's containers. Useful for debugging, monitoring progress, and understanding job behavior.",
    tags=["info", "avoid"],
    response_description="Job logs")
def job_logs(
    job_name: str,
    force_namespace: str = Query(None),
    pod_name: str = Query(None),
    tail: int = Query(100),
    api_key: str = Depends(verify_api_key)
):
    """
    Get job logs with the following parameters:
    
    - **job_name**: Name of the job
    - **force_namespace**: Optional namespace override
    - **pod_name**: Optional pod name
    - **tail**: Number of log lines to return
    """
    return fetch_job_logs(
        job_name=job_name,
        force_namespace=force_namespace,
        pod_name=pod_name,
        tail=tail
    )

@app.get("/fetch_job_templates",
    operation_id="fetch_job_templates",
    summary="Get available job templates",
    description="Retrieves a list of all available job templates that can be used to deploy workloads. Templates provide predefined configurations for frameworks.",
    tags=["info"],
    response_description="List of job templates")
def job_templates(api_key: str = Depends(verify_api_key)):
    return fetch_job_templates()

@app.get("/fetch_model_templates",
    operation_id="fetch_model_templates",
    summary="Get available model engines templates",
    description="Retrieves a list of all available model engine templates that can be used to deploy models. Templates provide predefined configurations for model engine frameworks.",
    tags=["info"],
    response_description="List of model engine templates")
def model_templates(api_key: str = Depends(verify_api_key)):
    return fetch_job_templates(type="model")

@app.get("/fetch_job_defaults",
    operation_id="fetch_job_defaults",
    summary="Get default values for a job or model engine template deployment",
    description="Retrieves the default values for a specific job or model engine template deployment. This helps users understand what parameters are required and what their default values are before deploying a job.",
    tags=["info"],
    response_description="Job and model engine default values")
def job_defaults(name: str, api_key: str = Depends(verify_api_key)):
    result = fetch_job_defaults(name=name)
    return result["defaults"]

@app.get("/fetch_job_metadata",
    operation_id="fetch_job_metadata",
    summary="Get metadata with information about a given job or model engine template deployment",
    description="Retrieves the metadata associated with a specific job or model engine template deployment. This helps users understand what the template can be used for.",
    tags=["info"],
    response_description="Job and model engine metadata values")
def job_metadata(name: str, api_key: str = Depends(verify_api_key)):
    result = fetch_job_defaults(name=name)
    return result["metadata"]

@app.get("/fetch_job_rules",
    operation_id="fetch_job_rules",
    summary="Get the rules associated with the use of a given job or model engine template",
    description="Retrieves the rules associated with a specific job or model engine template deployment. This helps users and AI agents determine if a given model engine template is adequate for the task.",
    tags=["info"],
    response_description="Job and model engine rules")
def job_rules(name: str, api_key: str = Depends(verify_api_key)):
    result = job_metadata(name=name)
    return result["template_rules"]

@app.get("/fetch_job_values_rules",
    operation_id="fetch_job_values_rules",
    summary="Get information on how to provide values to the parameters of a specific job or model engine template",
    description="Retrieves information necessary to fill up the values required to deploy a specific job or model engine template. This helps users and AI agents generate the values dictionary for a job or model engine template deployment.",
    tags=["info"],
    response_description="Job and model engine info for values")
def job_values_rules(name: str, api_key: str = Depends(verify_api_key)):
    result = job_metadata(name=name)
    return result["values_rules"]

@app.post("/deploy_job",
    operation_id="deploy_job",
    summary="Deploy a new job to the pool",
    description="Deploys a new job to the Kalavai pool using one of the available templates and configuration. The job will be scheduled on appropriate nodes based on resource availability and any specified target labels.",
    tags=["job_management"],
    response_description="Result of job deployment")
def job_deploy(request: DeployJobRequest, api_key: str = Depends(verify_api_key)):
    """
    Deploy a job with the following parameters:
    
    - **template_name**: Name of the job template
    - **values**: Job configuration values
    - **force_namespace**: Optional namespace override
    - **target_labels**: Optional target node labels
    """
    result = deploy_job(
        template_name=request.template_name,
        values_dict=request.values,
        force_namespace=request.force_namespace,
        target_labels=request.target_labels
    )
    return result

@app.post("/deploy_custom_job",
    operation_id="deploy_bustom_job",
    summary="Deploy a new custom job to the pool",
    description="Deploys a new job to the Kalavai pool using a specified template and configuration. The job will be scheduled on appropriate nodes based on resource availability and any specified target labels.",
    tags=["job_management"],
    response_description="Result of job deployment")
def custom_job_deploy(request: CustomDeployJobRequest, api_key: str = Depends(verify_api_key)):
    """
    Deploy a custom job with the following parameters:
    """
    result = deploy_test_job(
        template_str=request.template_str,
        values_dict=request.values,
        force_namespace=request.force_namespace,
        default_values=request.default_values,
        target_labels=request.target_labels
    )
    return result

@app.post("/delete_job",
    operation_id="delete_job",
    summary="Terminate and remove a job from the pool",
    description="Terminates a running job and removes it from the Kalavai pool. This will stop all containers associated with the job and free up the resources they were using.",
    tags=["job_management"],
    response_description="Result of job deletion")
def job_delete(request: DeleteJobRequest, api_key: str = Depends(verify_api_key)):
    """
    Delete a job with the following parameters:
    
    - **name**: Name of the job to delete
    - **force_namespace**: Optional namespace override
    """
    result = delete_job(
        name=request.name,
        force_namespace=request.force_namespace
    )
    return result

@app.get("/authenticate_user",
    operation_id="authenticate_user",
    summary="Authenticate a user with the Kalavai system",
    description="Authenticates a user against the Kalavai system, establishing their identity and permissions. This is required for accessing pool management features and deploying jobs.",
    tags=["info", "auth"],
    response_description="Authentication result")
def user_authenticate(user_id: str, api_key: str = Depends(verify_api_key)):
    """
    Authenticate user with the following parameters:
    
    - **user_id**: User identifier
    """
    result = authenticate_user(
        user_id=user_id
    )
    return result

@app.get("/load_user_session",
    operation_id="load_user_session",
    summary="Load current user session information",
    description="Retrieves information about the currently authenticated user's session, including their identity, permissions, and any active connections to pools.",
    tags=["info", "auth"],
    response_description="User session information")
def user_session(api_key: str = Depends(verify_api_key)):
    """Load the current user session"""
    result = load_user_session()
    return result

@app.get("/user_logout",
    operation_id="user_logout",
    summary="Log out the current user",
    description="Terminates the current user's session and clears authentication credentials. This should be called when the user is done using the system to ensure proper cleanup.",
    tags=["auth"],
    response_description="Logout result")
def logout_user():
    """Log out the current user"""
    result = user_logout()
    return result

@app.get("/is_connected",
    operation_id="is_connected",
    summary="Check if connected to a Kalavai pool",
    description="Verifies whether the current instance is connected to a Kalavai pool. Returns connection status and pool information if connected.",
    tags=["agent_management"],
    response_description="Connection status")
def pool_connected():
    """Check if connected to a pool"""
    result = is_connected()
    return result

@app.get("/is_agent_running",
    operation_id="is_agent_running",
    summary="Check if the Kalavai agent is running",
    description="Verifies whether the Kalavai agent service is currently running on this machine. The agent is responsible for managing pool connections and job execution.",
    tags=["agent_management"],
    response_description="Agent status")
def agent_running():
    """Check if the agent is running"""
    result = is_agent_running()
    return result

@app.get("/is_server",
    operation_id="is_server",
    summary="Check if running as a pool server",
    description="Determines whether this instance is running as a Kalavai pool server (coordinator) or as a client node. Server instances manage the pool while client instances contribute resources.",
    tags=["agent_management"],
    response_description="Server status")
def server():
    """Check if running as server"""
    result = is_server()
    return result

@app.post("/pause_agent",
    operation_id="pause_agent",
    summary="Pause the Kalavai agent service",
    description="Temporarily pauses the Kalavai agent, stopping it from accepting new jobs or participating in pool operations. Existing jobs will continue running until completion.",
    tags=["agent_management"],
    response_description="Result of pausing agent")
def agent_pause():
    """Pause the agent"""
    result = pause_agent()
    return result

@app.post("/resume_agent",
    operation_id="resume_agent",
    summary="Resume the Kalavai agent service",
    description="Resumes the previously paused Kalavai agent, allowing it to accept new jobs and participate in pool operations again.",
    tags=["agent_management"],
    response_description="Result of resuming agent")
def agent_resume():
    """Resume the agent"""
    result = resume_agent()
    return result

@app.get("/get_ip_addresses",
    operation_id="get_ip_addresses",
    summary="Get available IP addresses for pool configuration",
    description="Retrieves a list of available IP addresses that can be used for pool configuration. Optionally filters by subnet to help with network planning and pool setup.",
    tags=["agent_management"],
    response_description="List of IP addresses")
def ip_addresses(subnet: str = None, api_key: str = Depends(verify_api_key)):
    """
    Get IP addresses with the following parameters:
    
    - **subnet**: Optional subnet to filter by
    """
    result = get_ip_addresses(subnet=subnet)
    return result

@app.post("/add_node_labels",
    operation_id="add_node_labels",
    summary="Add custom labels to a compute node",
    description="Adds custom labels to a specific compute node in the pool. Labels can be used for job scheduling, resource allocation, and organizational purposes. Labels are key-value pairs that help categorize and identify nodes.",
    tags=["pool_management"],
    response_description="Result of adding labels")
def node_labels(request: NodeLabelsRequest, api_key: str = Depends(verify_api_key)):
    """
    Add node labels with the following parameters:
    
    - **node_name**: Name of the node
    - **labels**: Dictionary of labels to add
    """
    result = add_node_labels(
        node_name=request.node_name,
        labels=request.labels
    )
    return result

@app.get("/get_node_labels",
    operation_id="get_node_labels",
    summary="Get labels for specified compute nodes",
    description="Retrieves all labels associated with specified compute nodes in the pool. Labels provide metadata about nodes and can be used for filtering and scheduling decisions.",
    tags=["info"],
    response_description="Node labels")
def node_labels_get(nodes: Optional[List[str]] = Query(None), api_key: str = Depends(verify_api_key)):
    """
    Get node labels with the following parameters:
    
    - **nodes**: List of node names to get labels for
    """
    result = get_node_labels(
        node_names=nodes
    )
    return result

@app.get("/get_available_user_spaces",
    operation_id="get_available_user_spaces",
    summary="Get available user spaces in the pool",
    description="Get available user spaces in the pool.",
    tags=["info"],
    response_description="Available user spaces")
def get_available_user_spaces(api_key: str = Depends(verify_api_key)):
    """
    Get node labels with the following parameters:
    
    - **nodes**: List of node names to get labels for
    """
    if FORCED_USER_SPACE_NAME is not None:
        return [FORCED_USER_SPACE_NAME]
    else:
        return get_user_spaces()

@app.get("/get_user_space_quota",
    operation_id="get_user_space_quota",
    summary="Get resource quota for the user space",
    description="Get resource quota for the user space, including GPU, CPU and memory allowance.",
    tags=["info"],
    response_description="Available resource quota")
def get_user_space_quota(space_name: str=None, api_key: str = Depends(verify_api_key)):
    """
    Get resource quota for the user space with the following parameters:
    
    - **user_id**: name of the user space to get resource quota for
    """
    if FORCED_USER_SPACE_NAME is not None:
        return get_space_quota(space_name=FORCED_USER_SPACE_NAME)
    else:
        return get_space_quota(space_name=space_name)

### BUILD MCP WRAPPER ###
mcp = FastApiMCP(
    app,
    name="Protected MCP",
    #exclude_operations=[],
    exclude_tags=[
        "auth",
        "agent_management",
        "job_management",
        "pool_management",
        "avoid"
    ]
)
mcp.mount()
##########################


def run_api(host="0.0.0.0", port=8001, log_level="critical"):
    uvicorn.run(
        app,
        host=host,
        port=port,
        log_level=log_level
    )

if __name__ == "__main__":
    parser = ArgumentParser()
    parser.add_argument("--host", default="0.0.0.0")
    parser.add_argument("--port", default=8000, type=int)
    parser.add_argument("--log_level", default="debug")

    args = parser.parse_args()
    run_api(host=args.host, port=args.port, log_level=args.log_level)
    


import json, base64
import os
import uuid
import requests
from pathlib import Path
import shutil
import subprocess
import re

from jinja2 import Template

from rich.table import Table
import yaml

import kalavai_client
from kalavai_client.auth import KalavaiAuth
from kalavai_client.env import (
    SERVER_IP_KEY,
    DOCKER_COMPOSE_TEMPLATE,
    DEFAULT_CONTAINER_NAME,
    DEFAULT_FLANNEL_IFACE,
    DEFAULT_VPN_CONTAINER_NAME,
    USER_COMPOSE_FILE,
    USER_COOKIE,
    KALAVAI_USER_ID,
    FORCE_WATCHER_API_KEY_URL,
    FORCE_WATCHER_API_URL,
    USER_LOCAL_SERVER_FILE,
    user_path
)
from kalavai_client.api_models import TokenType


GITHUB_ORG = "kalavai-net"
GITHUB_REPO = "kalavai-client"
GITHUB_TEMPLATE_PATH = "templates"
USER_NODE_LABEL_KEY = "user_node_label"
CLUSTER_IP_KEY = "cluster_ip"
CLUSTER_TOKEN_KEY = "cluster_token"
NODE_NAME_KEY = "node_name"
PUBLIC_LOCATION_KEY = "public_location"
CLUSTER_NAME_KEY = "cluster_name"
AUTH_KEY = "watcher_admin_key"
USER_ID_KEY = "kalavai_user_id"
WRITE_AUTH_KEY = "watcher_write_key"
NODE_ROLE_LABEL = "kalavai/role"
USER_API_KEY = "user_api_key"
READONLY_AUTH_KEY = "watcher_readonly_key"
WATCHER_SERVICE_KEY = "watcher_service"
WATCHER_PORT_KEY = "watcher_port"
WATCHER_IMAGE_TAG_KEY = "watcher_image_tag"
KALAVAI_API_URL_KEY = "kalavai_api_url"
KALAVAI_API_KEY_KEY = "kalavai_api_key"
ADMIN_TOKEN_KEY = "admin_token"
USER_TOKEN_KEY = "user_token"
WORKER_TOKEN_KEY = "worker_token"
ENDPOINT_PORTS_KEY = "endpoint_ports"
TEMPLATE_ID_FIELD = "id_field"
TEMPLATE_ID_KEY = "deployment_id"
MANDATORY_TOKEN_FIELDS = [
    CLUSTER_IP_KEY,
    CLUSTER_TOKEN_KEY,
    CLUSTER_NAME_KEY,
    AUTH_KEY,
    WATCHER_SERVICE_KEY,
    PUBLIC_LOCATION_KEY
]
MANDATORY_POOLCONFIG_FIELDS = [
    SERVER_IP_KEY,
    AUTH_KEY,
    WATCHER_SERVICE_KEY,
    NODE_NAME_KEY,
    CLUSTER_NAME_KEY,
    PUBLIC_LOCATION_KEY
]

KALAVAI_AUTH = KalavaiAuth(
    auth_service_url="dummy_url",
    auth_service_key="dummy_key",
    user_cookie_file=USER_COOKIE
)


####### Methods to check OS compatibility ########
def check_gpu_drivers():
    value = run_cmd("nvidia-smi", hide_output=True)
    return len(value.decode("utf-8")) == 0

def get_max_gpus():
    try:
        has_gpus = check_gpu_drivers()
        if has_gpus:
            return len(
                [r for r in run_cmd("nvidia-smi -L").decode().split("\n") if len(r.strip())>0]
            )
        else:
            return 0
    except:
        return 0

def is_storage_compatible():
    """
    Raw method to determine if running node is compatible with PVC (longhorn)

    Exclude: WSL
    """
    try:
        import platform
        if "windows" in platform.system().lower():
            return True
        flagged = any([
            "microsoft" in run_cmd("cat /proc/version").decode().lower()
        ])
        return not flagged
    except:
        return False
################

def parse_key_value_pairs(input_str: str) -> dict:
    """Parse key=value pairs from a string into a dictionary.
    
    Args:
        input_str: String containing key=value pairs separated by commas
        
    Returns:
        Dictionary with parsed key-value pairs
        
    Raises:
        ValueError: If any pair is not in key=value format
    """
    if not input_str.strip():
        return {}
    
    result = {}
    for pair in input_str.split(','):
        pair = pair.strip()
        if not pair:
            continue
        if '=' not in pair:
            raise ValueError(f"Invalid key=value pair: '{pair}'. Expected format: key=value")
        key, value = pair.split('=', 1)
        result[key.strip()] = value.strip()
    return result

def extract_auth_token(headers):
    """
    Extract auth token. Valid headers:
        X-API-KEY: token
        X-API-Key: token
        Authorization: Bearer token
        authorization: Bearer token
    """
    #return headers.get("X-API-Key")
    bearer = None
    try:
        for header in ["Authorization", "authorization", "X-API-KEY", "X-API-Key"]:
            bearer = headers.get(header, None)
            if bearer is not None:
                break
        if bearer is not None and " " in bearer:
            return bearer.split()[-1]
        else:
            return bearer
    except Exception as e:
        return {"error": str(e)}


def generate_compose_config(
    role,
    node_name,
    mtu="",
    target_platform="amd64",
    write_to_file=True,
    node_ip_address="0.0.0.0",
    lb_ip_address=None,
    num_gpus=0,
    node_labels=None,
    pool_ip=None,
    vpn_token=None,
    pool_token=None,
    host_root_path=None,
    watcher_api_url=None,
    watcher_api_key=None,
    kalavai_api_port=49152,
    kalavai_api_key=None,
    kalavai_api_version=None
):
    if node_labels is not None:
        node_labels = " ".join([f"--node-label {key}={value}" for key, value in node_labels.items()])
    rand_suffix = uuid.uuid4().hex[:8]

    compose_values = {
        "target_platform": target_platform,
        "user_path": user_path(""),
        "service_name": DEFAULT_CONTAINER_NAME,
        "vpn": vpn_token is not None,
        "vpn_name": DEFAULT_VPN_CONTAINER_NAME,
        "mtu": mtu,
        "node_ip_address": node_ip_address,
        "load_balancer_ip_address": lb_ip_address if lb_ip_address is not None else "",
        "pool_ip": pool_ip,
        "pool_token": pool_token,
        "vpn_token": vpn_token,
        "node_name": node_name,
        "command": role,
        "storage_enabled": "True",
        "num_gpus": num_gpus,
        "user_local_path": os.path.expanduser("~"),
        "random_suffix": rand_suffix,
        "node_labels": node_labels,
        "flannel_iface": DEFAULT_FLANNEL_IFACE if vpn_token is not None else "",
        "user_id": load_user_id(),
        "host_root_path": host_root_path,
        "watcher_api_url": watcher_api_url,
        "watcher_api_key": watcher_api_key,
        "kalavai_api_port": kalavai_api_port,
        "kalavai_api_key": kalavai_api_key,
        "kalavai_api_image_version": kalavai_api_version if kalavai_api_version is not None else kalavai_client.__version__ 
    }
    # generate local config files
    compose_yaml = load_template(
        template_path=DOCKER_COMPOSE_TEMPLATE,
        values=compose_values)
    if write_to_file:
        with open(USER_COMPOSE_FILE, "w") as f:
            f.write(compose_yaml)
    return compose_yaml

def is_watcher_alive(server_creds, user_cookie, timeout=30):
    try:
        request_to_server(
            force_url=FORCE_WATCHER_API_URL,
            force_key=FORCE_WATCHER_API_KEY_URL,
            method="get",
            endpoint="/v1/health",
            data=None,
            server_creds=server_creds,
            user_cookie=user_cookie,
            timeout=timeout
        )
    except Exception as e:
        return False
    return True


def load_server_info(data_key, file):
    try:
        with open(file, "r") as f:
            return json.load(f)[data_key]
    except Exception as e:
        print(f"Warning: error when loading server info: {str(e)}")
        return None

def load_user_session():
    if KALAVAI_USER_ID is None:
        user_id = KALAVAI_AUTH.load_user_session()
    else:
        user_id = KALAVAI_USER_ID
    return user_id

def load_user_id():
    if KALAVAI_USER_ID is None:
        user_id = KALAVAI_AUTH.get_user_id()
    else:
        user_id = KALAVAI_USER_ID
    return user_id

def validate_poolconfig(poolconfig_file):
    if not Path(poolconfig_file).is_file():
        return False
    with open(poolconfig_file, "r") as f:
        data = json.load(f)
    for field in MANDATORY_POOLCONFIG_FIELDS:
        if field not in data:
            return False
    return True

def run_cmd(command, hide_output=False):
    try:
        import platform
        if "windows" in platform.system().lower():
            if hide_output:
                command = command + " > $nul 2>&1"
            return_value = subprocess.check_output(command, shell=True)
        else:
            if hide_output:
                command = command + " >/dev/null 2>&1"
            return_value = subprocess.check_output(command, shell=True, executable="/bin/bash")
        return return_value
    except OSError as error:
        return error # for exit code

def leave_vpn(container_name):
    try:
        vpns = json.loads(run_cmd(f"docker exec {container_name} netclient list").decode())
        left_vpns = [vpn['network'] for vpn in vpns]
        for vpn in left_vpns:
            run_cmd(f"docker exec {container_name} netclient leave {vpn}")
        return left_vpns
    except:
        return None
    
def has_api_details():
    """
    Check that the local machine has credentials that point to a
    remote or local kalavai API.

    Required when using the CLI
    """
    return all([cred is not None for cred in [
        load_server_info(data_key=KALAVAI_API_URL_KEY, file=USER_LOCAL_SERVER_FILE),
        load_server_info(data_key=KALAVAI_API_KEY_KEY, file=USER_LOCAL_SERVER_FILE)
    ]])

    
def request_to_api(
    method,
    endpoint,
    timeout=60,
    **kwargs
):
    """
    Calls to the Kalavai backend API
    
    Could be local or remote, reference in the server_creds file
    """
    api_url = load_server_info(data_key=KALAVAI_API_URL_KEY, file=USER_LOCAL_SERVER_FILE)
    api_key = load_server_info(data_key=KALAVAI_API_KEY_KEY, file=USER_LOCAL_SERVER_FILE)

    headers = {
        "X-API-KEY": api_key
    }

    response = requests.request(
        method=method,
        url=f"{api_url}{endpoint}",
        headers=headers,
        timeout=timeout,
        **kwargs
    )
    try:
        result = response.json()
        return result
    except Exception as e:
        raise ValueError(f"Error with HTTP request: {response.text}\n{str(e)}")


def request_to_server(
    method,
    endpoint,
    server_creds,
    data=None,
    params=None,
    force_url=None,
    force_key=None,
    user_cookie=None,
    timeout=60
):
    if force_url is None:
        service_url = load_server_info(data_key=WATCHER_SERVICE_KEY, file=server_creds)
    else:
        service_url = force_url
    
    if force_key is None:
        auth_key = load_server_info(data_key=AUTH_KEY, file=server_creds)
    else:
        auth_key = force_key

    headers = {
        "X-API-KEY": auth_key
    }

    headers["USER-KEY"] = load_server_info(data_key=USER_API_KEY, file=server_creds)
    user_id = load_user_id()
    if user_id is not None:
        headers["USER"] = user_id

    response = requests.request(
        method=method,
        url=f"http://{service_url}{endpoint}",
        json=data,
        params=params,
        headers=headers,
        timeout=timeout
    )
    try:
        result = response.json()
        return result
    except Exception as e:
        raise ValueError(f"Error with HTTP request: {response.text}\n{str(e)}")


def generate_table(columns, rows, end_sections=None):

    table = Table(show_header=True, header_style="bold white")
    [table.add_column(col, overflow="fold") for col in columns]
    for idx, row in enumerate(rows):
        table.add_row(*row, end_section=end_sections and idx in end_sections)

    return table

def store_server_info(
    server_ip,
    auth_key,
    watcher_service,
    file,
    node_name,
    cluster_name,
    readonly_auth_key=None,
    write_auth_key=None,
    public_location=None,
    user_api_key=None,
    kalavai_api_url="localhost:49152",
    kalavai_api_key=None,
    cluster_token=None,
):
    with open(file, "w") as f:
        json.dump({
            SERVER_IP_KEY: server_ip,
            AUTH_KEY: auth_key,
            READONLY_AUTH_KEY: readonly_auth_key,
            WRITE_AUTH_KEY: write_auth_key,
            WATCHER_SERVICE_KEY: watcher_service,
            NODE_NAME_KEY: node_name,
            CLUSTER_NAME_KEY: cluster_name,
            CLUSTER_TOKEN_KEY: cluster_token,
            PUBLIC_LOCATION_KEY: public_location,
            USER_API_KEY: user_api_key,
            KALAVAI_API_URL_KEY: kalavai_api_url,
            KALAVAI_API_KEY_KEY: kalavai_api_key
        }, f)
    return True

def populate_template(template_str, values_dict):
    return Template(template_str).render(values_dict)

def escape_field(text):
    return re.sub('[^0-9a-z]+', '-', text.lower())

def load_template(template_path, values, default_values_path=None, force_defaults=False):

    if not Path(template_path).exists():
        raise FileNotFoundError(f"{template_path} does not exist")
    with open(template_path, 'r') as f:
        yaml_template = "".join(f.readlines())
    
    # substitute missing values with defaults
    if default_values_path is not None:
        with open(default_values_path, 'r') as f:
            default_values = yaml.safe_load(f)
        for default in default_values:
            if default["name"] == TEMPLATE_ID_FIELD:
                if default["default"] not in values:
                    raise ValueError(f"Key value '{default['default']}' missing from values")
                values[TEMPLATE_ID_KEY] = escape_field(values[default["default"]])
                continue
            if force_defaults or default["name"] not in values:
                values[default['name']] = default['default']
        
    return populate_template(template_str=yaml_template, values_dict=values)


def user_confirm(question: str, options: list, multiple: bool=False) -> int:
    try:
        print(question)
        [print(f"{idx}) {option}") for idx, option in enumerate(options)]
        reply = str(input())
        if multiple:
            if "," in reply:
                selection = reply.split(",")
                cast_selection = [int(s) for s in selection]
                if all([s >= 0 and s < len(options) for s in cast_selection]):
                    return cast_selection
            else:
                reply = int(reply)
                if reply >= 0 and reply <= len(options):
                    return [reply]
        else:
            reply = int(reply)
            if reply >= 0 and reply <= len(options):
                return reply

    except:
        return None
    return None

def generate_join_token(cluster_ip, cluster_token, cluster_name, auth_key, watcher_service, public_location):
    data = {
        CLUSTER_IP_KEY: cluster_ip,
        CLUSTER_NAME_KEY: cluster_name,
        CLUSTER_TOKEN_KEY: cluster_token,
        AUTH_KEY: auth_key,
        WATCHER_SERVICE_KEY: watcher_service,
        PUBLIC_LOCATION_KEY: public_location
    }
    return encode_dict(data=data)

def encode_dict(data: dict):
    base64_str = base64.b64encode(json.dumps(data).encode())
    return base64_str.decode()

def decode_dict(str_data: str):
    return json.loads(base64.b64decode(str_data.encode()))

def safe_remove(filepath, force=True):
    if not os.path.exists(filepath):
        return
    try:
        if os.path.isfile(filepath):
            os.remove(filepath)
        if os.path.isdir(filepath):
            shutil.rmtree(filepath)
        return
    except:
        pass

    if force:
        try:
            run_cmd(f"rm -rf {filepath}")
        except:
            pass



import yaml
import time
from collections import defaultdict
import uuid
import socket
import ipaddress
import netifaces as ni
from urllib.parse import urlparse

from kalavai_client.cluster import CLUSTER
from kalavai_client.utils import (
    NODE_ROLE_LABEL,
    check_gpu_drivers,
    generate_join_token,
    load_user_id,
    request_to_server,
    load_server_info,
    decode_dict,
    generate_compose_config,
    store_server_info,
    is_watcher_alive,
    run_cmd,
    leave_vpn,
    safe_remove,
    load_template,
    is_storage_compatible,
    get_max_gpus,
    NODE_NAME_KEY,
    MANDATORY_TOKEN_FIELDS,
    PUBLIC_LOCATION_KEY,
    CLUSTER_IP_KEY,
    CLUSTER_NAME_KEY,
    AUTH_KEY,
    USER_ID_KEY,
    WATCHER_SERVICE_KEY,
    CLUSTER_TOKEN_KEY,
    READONLY_AUTH_KEY,
    WRITE_AUTH_KEY,
    WATCHER_PORT_KEY,
    WATCHER_SERVICE_KEY,
    WATCHER_IMAGE_TAG_KEY,
    USER_NODE_LABEL_KEY,
    KALAVAI_API_URL_KEY,
    KALAVAI_API_KEY_KEY,
    KALAVAI_AUTH
)
from kalavai_client.env import (
    USER_COOKIE,
    USER_LOCAL_SERVER_FILE,
    TEMPLATE_LABEL,
    SERVER_IP_KEY,
    USER_COMPOSE_FILE,
    DEFAULT_VPN_CONTAINER_NAME,
    CONTAINER_HOST_PATH,
    USER_VPN_COMPOSE_FILE,
    USER_HELM_APPS_FILE,
    USER_KUBECONFIG_FILE,
    USER_TEMPLATES_FOLDER,
    USER_WORKSPACE_TEMPLATE,
    DEFAULT_USER_WORKSPACE_VALUES,
    STORAGE_CLASS_LABEL,
    USER_NODE_LABEL,
    DEFAULT_WATCHER_PORT,
    HELM_APPS_FILE,
    POOL_CONFIG_TEMPLATE,
    FORBIDEDEN_IPS,
    DEFAULT_POOL_CONFIG_TEMPLATE,
    FORCE_WATCHER_API_URL,
    FORCE_WATCHER_API_KEY_URL
)
from kalavai_client.api_models import (
    GPU,
    Job,
    TokenType,
    DeviceStatus
)


def set_schedulable(schedulable, node_names):
    """
    Delete job in the cluster
    """
    # deploy template with kube-watcher
    data = {
        "schedulable": str(schedulable),
        "node_names": node_names
    }
    try:
        res = request_to_server(
            force_url=FORCE_WATCHER_API_URL,
            force_key=FORCE_WATCHER_API_KEY_URL,
            method="post",
            endpoint="/v1/set_node_schedulable",
            data=data,
            server_creds=USER_LOCAL_SERVER_FILE,
            user_cookie=USER_COOKIE
        )
        if res is not None and "detail" in res:
            return {"error": res["detail"]}
        else:
            return {"result": res}
    except Exception as e:
        return {"error": f"Error when connecting to kalavai service: {str(e)}"}

def init_user_workspace(user_id=None, node_name=None, force_namespace=None):
    
    # load template config and populate with values
    sidecar_template_yaml = load_template(
        template_path=USER_WORKSPACE_TEMPLATE,
        values={},
        default_values_path=DEFAULT_USER_WORKSPACE_VALUES)

    try:
        data = {"config": sidecar_template_yaml}
        if force_namespace is not None:
            data["force_namespace"] = force_namespace
        if user_id is not None:
            data["user_id"] = user_id
        if node_name is not None:
            data["node_name"] = node_name
        result = request_to_server(
            force_url=FORCE_WATCHER_API_URL,
            force_key=FORCE_WATCHER_API_KEY_URL,
            method="post",
            endpoint="/v1/create_user_space",
            data=data,
            server_creds=USER_LOCAL_SERVER_FILE,
            user_cookie=USER_COOKIE
        )
        return {"success"}
    except Exception as e:
        return {"error": f"Error when connecting to kalavai service: {str(e)}"}

def check_seed_compatibility():
    """Check required packages to start pools"""
    logs = []
    # docker
    try:
        run_cmd("docker ps", hide_output=True)
    except:
        logs.append("[red]Docker not installed. Install instructions:\n")
        logs.append("   Linux: https://docs.docker.com/engine/install/\n")
        logs.append("   Windows/MacOS: https://docs.docker.com/desktop/\n")
    
    return {"issues": logs}

def check_worker_compatibility():
    """Check required packages to join pools"""
    logs = []
    # docker
    try:
        run_cmd("docker ps", hide_output=True)
    except:
        logs.append("[red]Docker not installed. Install instructions:\n")
        logs.append("   Linux: https://docs.docker.com/engine/install/\n")
        logs.append("   Windows/MacOS: https://docs.docker.com/desktop/\n")
    
    return {"issues": logs}

def get_ip_addresses(subnet=None):
    ips = []
    retry = 3
    while len(ips) == 0:
        for iface in ni.interfaces():
            try:
                ip = ni.ifaddresses(iface)[ni.AF_INET][0]['addr']
                if ip in FORBIDEDEN_IPS:
                    continue
                if subnet is None or ipaddress.ip_address(ip) in ipaddress.ip_network(subnet):
                    ips.append(ip)
            except:
                pass
        time.sleep(2)
        retry -= 1
        if retry < 0:
            raise ValueError(f"No IPs available on subnet {subnet}")
    return ips

def fetch_resources(node_names: list[str]=None):
    data = {}
    if node_names is not None:
        data["node_names"] = node_names
    try:
        total = request_to_server(
            force_url=FORCE_WATCHER_API_URL,
            force_key=FORCE_WATCHER_API_KEY_URL,
            method="post",
            endpoint="/v1/get_cluster_total_resources",
            data=data,
            server_creds=USER_LOCAL_SERVER_FILE,
            user_cookie=USER_COOKIE
        )
        available = request_to_server(
            force_url=FORCE_WATCHER_API_URL,
            force_key=FORCE_WATCHER_API_KEY_URL,
            method="post",
            endpoint="/v1/get_cluster_available_resources",
            data=data,
            server_creds=USER_LOCAL_SERVER_FILE,
            user_cookie=USER_COOKIE
        )
    except Exception as e:
        return {"error": str(e)}
        
    return {"total": total, "available": available}

def get_user_spaces():

    try:
        data = request_to_server(
            force_url=FORCE_WATCHER_API_URL,
            force_key=FORCE_WATCHER_API_KEY_URL,
            method="get",
            endpoint="/v1/get_available_user_spaces",
            server_creds=USER_LOCAL_SERVER_FILE,
            user_cookie=USER_COOKIE
        )
        return data
    except Exception as e:
        return {"error": str(e)}
    
def get_space_quota(space_name):

    try:
        data = request_to_server(
            force_url=FORCE_WATCHER_API_URL,
            force_key=FORCE_WATCHER_API_KEY_URL,
            method="get",
            endpoint="/v1/get_space_quota",
            params={"user_id": space_name},
            server_creds=USER_LOCAL_SERVER_FILE,
            user_cookie=USER_COOKIE
        )
        return data
    except Exception as e:
        return {"error": str(e)}
    

def fetch_job_defaults(name):
    data = {
        "template_name": name
    }
    try:
        metadata = request_to_server(
            force_url=FORCE_WATCHER_API_URL,
            force_key=FORCE_WATCHER_API_KEY_URL,
            method="get",
            endpoint="/v1/job_defaults",
            params=data,
            server_creds=USER_LOCAL_SERVER_FILE,
            user_cookie=USER_COOKIE
        )
        return metadata
    except Exception as e:
        return {"error": str(e)}
    
def fetch_job_templates(type: str=None):
    data = None
    if type is not None:
        data = {"type": type}
    try:
        templates = request_to_server(
            force_url=FORCE_WATCHER_API_URL,
            force_key=FORCE_WATCHER_API_KEY_URL,
            method="get",
            endpoint="/v1/get_job_templates",
            server_creds=USER_LOCAL_SERVER_FILE,
            data=None,
            params=data,
            user_cookie=USER_COOKIE
        )
        return templates
    except Exception as e:
        return {"error": str(e)}

def fetch_job_names():
    data_groups = [
        {
            "group": "batch.volcano.sh",
            "api_version": "v1alpha1",
            "plural": "jobs"
        },
        {
            "group": "ray.io",
            "api_version": "v1",
            "plural": "rayclusters"
        }
    ]
    try:
        all_jobs = []
        for data in data_groups:
            jobs = request_to_server(
                force_url=FORCE_WATCHER_API_URL,
                force_key=FORCE_WATCHER_API_KEY_URL,
                method="post",
                endpoint="/v1/get_objects_of_type",
                data=data,
                server_creds=USER_LOCAL_SERVER_FILE,
                user_cookie=USER_COOKIE
            )
            for ns, ds in jobs.items():
                all_jobs.extend([Job(owner=ns, name=d["metadata"]["labels"][TEMPLATE_LABEL]) for d in ds["items"]])
            
    except Exception as e:
        return {"error": str(e)}
    
    return all_jobs  

def fetch_job_details(force_namespace=None):
    """Get jobs overview details (status and services)"""
    job_details = []
    # fetch all details at once
    data = {"labels": [TEMPLATE_LABEL]}
    if force_namespace is not None:
        data["force_namespace"] = force_namespace
    result = request_to_server(
        force_url=FORCE_WATCHER_API_URL,
        force_key=FORCE_WATCHER_API_KEY_URL,
        method="post",
        endpoint="/v1/get_jobs_overview",
        data=data,
        server_creds=USER_LOCAL_SERVER_FILE,
        user_cookie=USER_COOKIE)

    for namespace, deployments in result.items():
        """deployments --> "matched_label_value": {"pods": [], "services": []} }"""
        for job_name, job in deployments.items():
            workers_status = defaultdict(int)
            workers = ""
            restart_counts = 0
            host_nodes = set()
            # parse pods
            if "pods" in job and job["pods"] is not None:
                for name, values in job["pods"].items():
                    if "conditions" in values and values["conditions"] is not None:
                        restart_counts = sum([c["restart_count"] for c in values["conditions"]])
                    workers_status[values["status"]] += 1
                    # get nodes involved in deployment (needs kubewatcher)
                    if "node_name" in values and values["node_name"] is not None:
                        host_nodes.add(values["node_name"])
                    workers = "\n".join([f"{k}: {v}" for k, v in workers_status.items()])
                    if restart_counts > 0:
                        workers += f"\n({restart_counts} restart)"
            else:
                print("Skip pods")
            # parse services
            node_ports = []
            if "services" in job and job["services"] is not None:
                for name, values in job["services"].items():
                    node_ports.extend(
                        [f"{port['node_port']}" for port in values["ports"]]
                    )
            else:
                print("Skip service")
            endpoint_address = urlparse(load_server_info(data_key=KALAVAI_API_URL_KEY, file=USER_LOCAL_SERVER_FILE)).hostname
            urls = [f"http://{endpoint_address}:{node_port}" for node_port in node_ports]
            if "Ready" in workers_status and len(workers_status) == 1:
                status = "running"
            elif any([st in workers_status for st in ["Failed"]]):
                status = "error"
            elif any([st in workers_status for st in ["Pending"]]) or len(workers_status) == 0:
                status = "pending"
            elif any([st in workers_status for st in ["Succeeded", "Completed"]]):
                status = "completed"
            else:
                status = "working"
            job_details.append(
                Job(owner=namespace,
                    name=job_name,
                    workers=workers,
                    endpoint="\n".join(urls),
                    status=str(status),
                    host_nodes="\n".join(host_nodes))
            )
    return job_details

def deploy_job(template_name, values_dict, force_namespace=None, target_labels=None):

    # deploy template with kube-watcher
    data = {
        "template": template_name,
        "template_values": values_dict
    }
    if force_namespace is not None:
        data["force_namespace"] = force_namespace
    if target_labels is not None:
        data["target_labels"] = target_labels

    try:
        result = request_to_server(
            force_url=FORCE_WATCHER_API_URL,
            force_key=FORCE_WATCHER_API_KEY_URL,
            method="post",
            endpoint="/v1/deploy_job",
            data=data,
            server_creds=USER_LOCAL_SERVER_FILE,
            user_cookie=USER_COOKIE
        )
        return result
    except Exception as e:
        return {"error": str(e)}  
    
def deploy_test_job(template_str, values_dict, default_values, target_labels=None, force_namespace=None):
    
    # submit custom deployment
    data = {
        "template": template_str,
        "template_values": values_dict,
        "default_values": default_values,
        "target_labels": target_labels
    }
    if force_namespace is not None:
        data["force_namespace"] = force_namespace

    try:
        result = request_to_server(
            force_url=FORCE_WATCHER_API_URL,
            force_key=FORCE_WATCHER_API_KEY_URL,
            method="post",
            endpoint="/v1/deploy_custom_job",
            data=data,
            server_creds=USER_LOCAL_SERVER_FILE,
            user_cookie=USER_COOKIE
        )
        return result
    except Exception as e:
        return {"error": str(e)}
    
def delete_job(name, force_namespace=None):
    data = {
        "label": TEMPLATE_LABEL, # this ensures that both lws template and services are deleted
        "value": name
    }
    if force_namespace is not None:
        data["force_namespace"] = force_namespace
    try:
        result = request_to_server(
            force_url=FORCE_WATCHER_API_URL,
            force_key=FORCE_WATCHER_API_KEY_URL,
            method="post",
            endpoint="/v1/delete_labeled_resources",
            data=data,
            server_creds=USER_LOCAL_SERVER_FILE,
            user_cookie=USER_COOKIE
        )
        return result
    except Exception as e:
        return {"error": str(e)}

def fetch_devices():
    """Load devices status info for all hosts"""
    try:
        data = request_to_server(
            force_url=FORCE_WATCHER_API_URL,
            force_key=FORCE_WATCHER_API_KEY_URL,
            method="get",
            endpoint="/v1/get_nodes",
            data={},
            server_creds=USER_LOCAL_SERVER_FILE,
            user_cookie=USER_COOKIE
        )
        devices = []
        for node, status in data.items():
            devices.append(
                DeviceStatus(
                    name=node,
                    memory_pressure=status["MemoryPressure"],
                    disk_pressure=status["DiskPressure"],
                    pid_pressure=status["PIDPressure"],
                    ready=status["Ready"],
                    unschedulable=status["unschedulable"]
                )
            )
        return devices

    except Exception as e:
        return {"error": str(e)}

def fetch_job_logs(job_name, force_namespace=None, pod_name=None, tail=100):
    return fetch_pod_logs(
        label_key=TEMPLATE_LABEL,
        label_value=job_name,
        pod_name=pod_name,
        force_namespace=force_namespace,
        tail=tail
    )

def fetch_pod_logs(label_key, label_value, force_namespace=None, pod_name=None, tail=100):
    data = {
        "label": label_key,
        "value": label_value,
        "tail_lines": tail
    }
    if force_namespace is not None:
        data["force_namespace"] = force_namespace
    try:
        # send tail as parameter (fetch only last _tail_ lines)
        all_logs = request_to_server(
            force_url=FORCE_WATCHER_API_URL,
            force_key=FORCE_WATCHER_API_KEY_URL,
            method="post",
            endpoint="/v1/get_job_details",
            data=data,
            server_creds=USER_LOCAL_SERVER_FILE,
            user_cookie=USER_COOKIE
        )
        return {pod: info for pod, info in all_logs.items() if pod_name is None or pod_name == pod}

    except Exception as e:
        return {"error": str(e)}
    

def load_gpu_models():
    data = request_to_server(
        force_url=FORCE_WATCHER_API_URL,
        force_key=FORCE_WATCHER_API_KEY_URL,
        method="post",
        endpoint="/v1/get_node_gpus",
        data={},
        server_creds=USER_LOCAL_SERVER_FILE,
        user_cookie=USER_COOKIE
    )
    return data.items()

def fetch_gpus(available=False):
    try:
        data = load_gpu_models()
        all_gpus = []
        for node, gpus in data:
            row_gpus = []
            for gpu in gpus["gpus"]:
                status = gpu["ready"] if "ready" in gpu else True
                if available and not status:
                    continue
                row_gpus.append( (f"{gpu['model']} ({gpu['memory']} vRAM)", str(status)))
            if len(row_gpus) > 0:
                models, statuses = zip(*row_gpus)
                #rows.append([node, "\n".join(statuses), "\n".join(models), str(gpus["available"]), str(gpus["capacity"])])
                all_gpus.extend([
                    GPU(
                        node=node,
                        ready=status,
                        model=model,
                        available=gpus["available"],
                        total=gpus["capacity"]
                    ) for model, status in zip(models, statuses)
                ])
        return all_gpus

    except Exception as e:
        return {"error": str(e)}
    
def authenticate_user(user_key=None):
    if user_key is None:
        KALAVAI_AUTH.save_auth(user_key)
    
    return KALAVAI_AUTH.load_user_session()

def load_user_session():
    return KALAVAI_AUTH.load_user_session()

def user_logout():
    return KALAVAI_AUTH.clear_auth()

def check_token(token, public=False):
    try:
        data = decode_dict(token)
        for field in MANDATORY_TOKEN_FIELDS:
            assert field in data
        if public:
            if data[PUBLIC_LOCATION_KEY] is None:
                raise ValueError("Token is not valid for public pools. Did you start the cluster with a public_location?")
        return {"status": True, "data": data}
    except Exception as e:
        return {"error": str(e)}
    
def delete_nodes(nodes):
    data = {
        "node_names": nodes
    }
    try:
        result = request_to_server(
            force_url=FORCE_WATCHER_API_URL,
            force_key=FORCE_WATCHER_API_KEY_URL,
            method="post",
            endpoint="/v1/delete_nodes",
            data=data,
            server_creds=USER_LOCAL_SERVER_FILE,
            user_cookie=USER_COOKIE
        )
        if result is None or result is True:
            return {"success": nodes}
        else:
            return {"error": result}
    except Exception as e:
        return {"error": f"Error when removing nodes {nodes}: {str(e)}"}

def cordon_nodes(nodes):
    return set_schedulable(schedulable=False, node_names=nodes)

def uncordon_nodes(nodes):
    return set_schedulable(schedulable=True, node_names=nodes)

def attach_to_pool(token, node_name=None):
    if node_name is None:
        node_name = socket.gethostname()
    
    # check token
    valid = check_token(token=token)
    if "error" in valid:
        return {"error": f"Invalid token: {valid}"}

    try:
        data = decode_dict(token)
        kalavai_seed_ip = data[CLUSTER_IP_KEY]
        cluster_name = data[CLUSTER_NAME_KEY]
        auth_key = data[AUTH_KEY]
        watcher_service = data[WATCHER_SERVICE_KEY]
        public_location = data[PUBLIC_LOCATION_KEY]
    except Exception as e:
        return {"error": f"Invalid token. {str(e)}"} 
    
    # local agent join
    # 1. Generate local cache files
    # Generate docker compose recipe
    generate_compose_config(
        role="",
        vpn_token=public_location,
        node_name=node_name)
    
    store_server_info(
        server_ip=kalavai_seed_ip,
        auth_key=auth_key,
        file=USER_LOCAL_SERVER_FILE,
        watcher_service=watcher_service,
        node_name=node_name,
        cluster_name=cluster_name,
        public_location=public_location,
        user_api_key=None)
    
    run_cmd(f"docker compose -f {USER_COMPOSE_FILE} up -d")
    # ensure we are connected
    while True:
        time.sleep(30)
        if is_watcher_alive(server_creds=USER_LOCAL_SERVER_FILE, user_cookie=USER_COOKIE):
            break

    return cluster_name

def generate_worker_package(
        target_platform="amd64",
        num_gpus=0,
        node_name=None,
        ip_address="0.0.0.0",
        storage_compatible=True,
        mode=TokenType.WORKER
):
    # get pool data from token  
    token = get_pool_token(mode=mode)
    if "error" in token:
        return {"error": f"[red]Error when getting pool token: {token['error']}"}
    
    if node_name is None:
        node_name = f"worker-{uuid.uuid4().hex[:6]}"
    
    # parse pool data
    try:
        data = decode_dict(token["token"])
        kalavai_seed_ip = data[CLUSTER_IP_KEY]
        kalavai_token = data[CLUSTER_TOKEN_KEY]
        public_location = data[PUBLIC_LOCATION_KEY]
    except Exception as e:
        return {"error": f"Invalid token. {str(e)}"} 
    
    # join private network if provided
    node_labels = {
        STORAGE_CLASS_LABEL: storage_compatible,
        NODE_ROLE_LABEL: "worker"
    }
    # Generate docker compose recipe
    compose = generate_compose_config(
        target_platform=target_platform,
        write_to_file=False,
        role="agent",
        node_ip_address=ip_address,
        pool_ip=f"https://{kalavai_seed_ip}:6443",
        pool_token=kalavai_token,
        num_gpus=num_gpus,
        vpn_token=public_location,
        node_name=node_name,
        node_labels=node_labels)
    
    return compose


def join_pool(
        token,
        num_gpus=None,
        node_name=None,
        ip_address=None,
        target_platform="amd64",
        kalavai_api_version=None,
        mtu="",
        node_labels={},
        is_seed=False
):
    compatibility = check_worker_compatibility()
    if len(compatibility["issues"]) > 0:
        return {"error": compatibility["issues"]}

    if num_gpus is None:
        num_gpus = get_max_gpus()

    if node_name is None:
        node_name = socket.gethostname()
    
    # check token
    valid = check_token(token=token)
    if "error" in valid:
        return {"error": f"Invalid token: {valid}"}

    try:
        data = decode_dict(token)
        kalavai_seed_ip = data[CLUSTER_IP_KEY]
        kalavai_token = data[CLUSTER_TOKEN_KEY]
        cluster_name = data[CLUSTER_NAME_KEY]
        auth_key = data[AUTH_KEY]
        watcher_service = data[WATCHER_SERVICE_KEY]
        public_location = data[PUBLIC_LOCATION_KEY]
        kalavai_api_port = 49152
    except Exception as e:
        return {"error": f"Invalid token. {str(e)}"} 
    
    # join private network if provided
    node_labels = {
        **node_labels,
        STORAGE_CLASS_LABEL: is_storage_compatible(),
        NODE_ROLE_LABEL: "worker" if not is_seed else "server"
    }  
    # local agent join
    # Generate docker compose recipe
    generate_compose_config(
        target_platform=target_platform,
        role="agent" if not is_seed else "seed",
        node_ip_address=ip_address,
        pool_ip=f"https://{kalavai_seed_ip}:6443",
        pool_token=kalavai_token,
        num_gpus=num_gpus,
        vpn_token=public_location,
        mtu=mtu,
        node_name=node_name,
        node_labels=node_labels,
        watcher_api_url=watcher_service,
        watcher_api_key=auth_key,
        kalavai_api_key=auth_key,
        kalavai_api_port=kalavai_api_port,
        kalavai_api_version=kalavai_api_version)
    
    store_server_info(
        server_ip=kalavai_seed_ip,
        auth_key=auth_key,
        file=USER_LOCAL_SERVER_FILE,
        watcher_service=watcher_service,
        node_name=node_name,
        cluster_name=cluster_name,
        public_location=public_location,
        user_api_key=None,
        kalavai_api_url=f"http://localhost:{kalavai_api_port}",
        kalavai_api_key=auth_key)
    
    try:
        CLUSTER.start_worker_node()
    except Exception as e:
        return {"error": f"Error connecting to {cluster_name} @ {kalavai_seed_ip}. Check with the admin if the token is still valid."}

    # ensure we are connected
    while True:
        time.sleep(30)
        if is_watcher_alive(server_creds=USER_LOCAL_SERVER_FILE, user_cookie=USER_COOKIE):
            break
    
    # check the node has connected successfully
    try:
        while not CLUSTER.is_agent_running():
            time.sleep(30)
    except KeyboardInterrupt:
        return {"error": "Installation aborted. Leaving pool."}
    
    result = init_user_workspace(
        user_id=load_user_id(),
        node_name=node_name)
    if "error" in result:
        return {"error": f"Error when creating user workspace: {result}"}
    
    return cluster_name

def create_pool(
    cluster_name: str=None,
    ip_address: str=None,
    lb_ip_address: str=None,
    location: str=None,
    target_platform: str="amd64",
    watcher_image_tag: str=None,
    pool_config_file: str=None,
    kalavai_api_version: str=None,
    num_gpus: int=-1,
    node_name: str=None,
    mtu: str="",
    apps: list=[],
    node_labels: dict={}
):

    if not check_seed_compatibility():
        return {"error": "Requirements failed"}
    
    if pool_config_file is None:
        pool_config_file = DEFAULT_POOL_CONFIG_TEMPLATE

    if node_name is None:
        node_name = socket.gethostname()

    user_id = load_user_id()
    
    node_labels = {
        **node_labels,
        STORAGE_CLASS_LABEL: is_storage_compatible(),
        NODE_ROLE_LABEL: "server"
    }
        
    if num_gpus < 0:
        num_gpus = get_max_gpus()

    # load values from pool config
    with open(pool_config_file, "r") as f:
        config_values = yaml.safe_load(f)
    # use default values if not provided
    try:
        watcher_image_tag = config_values["server"]["watcher_image_tag"] if watcher_image_tag is None else watcher_image_tag
        cluster_name = config_values["server"]["name"] if cluster_name is None else cluster_name
        ip_address = config_values["server"]["ip_address"] if ip_address is None else ip_address
        location = config_values["server"]["location"] if location is None else location
        target_platform = config_values["server"]["platform"] if target_platform is None else target_platform
        mtu = config_values["server"]["mtu"] if mtu == "" or mtu is None else mtu
        app_values = config_values["core"]
        kalavai_api_version = app_values["kalavai_api_version"] if kalavai_api_version is None else kalavai_api_version
        post_config_values = config_values["pool"]
        deploy_apps = {
            f"deploy_{app}": True for app in config_values["core"]["deploy"]
        }
        for app in apps:
            deploy_apps[f"deploy_{app}"] = True
    except Exception as e:
        return {"error": f"Error when loading pool config. Missing format? {str(e)}"}

    # Generate docker compose recipe
    auth_key = user_id if user_id is not None else str(uuid.uuid4())
    write_auth_key = str(uuid.uuid4())
    readonly_auth_key = str(uuid.uuid4())
    watcher_service = f"{ip_address}:{DEFAULT_WATCHER_PORT}"
    kalavai_api_port = 49152

    if ip_address is None:
        ip_address = "0.0.0.0"
    generate_compose_config(
        target_platform=target_platform,
        role="server",
        vpn_token=location,
        node_ip_address=ip_address,
        lb_ip_address=lb_ip_address,
        num_gpus=num_gpus,
        node_name=node_name,
        node_labels=node_labels,
        mtu=mtu,
        host_root_path=config_values["server"]["host_root_path"],
        watcher_api_key=auth_key,
        watcher_api_url=watcher_service,
        kalavai_api_port=kalavai_api_port,
        kalavai_api_key=auth_key,
        kalavai_api_version=kalavai_api_version
    )

    # start server
    CLUSTER.start_seed_node()
    while not CLUSTER.is_agent_running():
        time.sleep(10)
    
    
    # select IP address (for external discovery)
    if ip_address is None or location is not None:
        # load VPN ip
        ip_address = None
        while ip_address is None or len(ip_address) == 0:
            ip_address = CLUSTER.get_vpn_ip()
            time.sleep(10)

    # populate local cred files
    values = {
        #CLUSTER_NAME_KEY: cluster_name,
        CLUSTER_IP_KEY: ip_address if lb_ip_address is None else lb_ip_address,
        USER_ID_KEY: user_id if user_id is not None else "",
        AUTH_KEY: auth_key,
        READONLY_AUTH_KEY: readonly_auth_key,
        WRITE_AUTH_KEY: write_auth_key,
        WATCHER_PORT_KEY: DEFAULT_WATCHER_PORT,
        WATCHER_SERVICE_KEY: watcher_service,
        USER_NODE_LABEL_KEY: USER_NODE_LABEL,
        WATCHER_IMAGE_TAG_KEY: watcher_image_tag,
    }

    store_server_info(
        server_ip=ip_address if lb_ip_address is None else lb_ip_address,
        auth_key=auth_key,
        readonly_auth_key=readonly_auth_key,
        write_auth_key=write_auth_key,
        file=USER_LOCAL_SERVER_FILE,
        watcher_service=watcher_service,
        node_name=node_name,
        cluster_name=cluster_name,
        public_location=location,
        user_api_key=None,
        kalavai_api_url=f"http://localhost:{kalavai_api_port}",
        kalavai_api_key=auth_key,
        cluster_token=CLUSTER.get_cluster_token())
    
    # Generate helmfile recipe
    helm_yaml = load_template(
        template_path=HELM_APPS_FILE,
        values={**values, **deploy_apps, **app_values},
        #default_values_path=app_values,
        force_defaults=True)
    with open(USER_HELM_APPS_FILE, "w") as f:
        f.write(helm_yaml)
    
    # set template values in helmfile
    try:
        CLUSTER.update_dependencies(
            dependencies_file=USER_HELM_APPS_FILE
        )
    except Exception as e:
        return {"error": f"Error when updating dependencies: {str(e)}"}
    
    # wait until the server is ready to create objects
    while True:
        time.sleep(30)
        if is_watcher_alive(server_creds=USER_LOCAL_SERVER_FILE, user_cookie=USER_COOKIE):
            break

    result = pool_init(config_values=post_config_values)
    if "error" in result or ("failed" in result and len(result['failed']) > 0):
        return {"error": f"Error when initialising pool: {result}"}
    # init default namespace
    init_user_workspace(
        user_id=user_id,
        node_name=node_name,
        force_namespace="default")
    # if only_registered_users:
    #     # init user namespace
    #     init_user_workspace(
    #         user_id=user_id)

    if "error" in result:
        return {"warning": result["error"]}
    if "warning" in result:
        return {"warning": result["warning"]}
    
    return {"success"}

def update_pool(debug=True):
    try:
        CLUSTER.validate_cluster()
    except Exception as e:
        return {"error": f"Problems with your pool: {str(e)}"}
    
    if not CLUSTER.is_seed_node():
        return {"error": "You can only update a pool from the seed node."}
    
    # update dependencies
    try:
        CLUSTER.update_dependencies(debug=debug)
        return {"success": "Pool updating. Expect some downtime on core services"}
    except Exception as e:
        return {"error": f"[red]Error when updating pool: {str(e)}"}

def get_pool_credentials():
    return {
        KALAVAI_API_URL_KEY: load_server_info(data_key=KALAVAI_API_URL_KEY, file=USER_LOCAL_SERVER_FILE),
        KALAVAI_API_KEY_KEY: load_server_info(data_key=KALAVAI_API_KEY_KEY, file=USER_LOCAL_SERVER_FILE)
    }

def get_pool_token(mode: TokenType):

    try:
        match mode:
            case TokenType.ADMIN:
                auth_key = load_server_info(data_key=AUTH_KEY, file=USER_LOCAL_SERVER_FILE)
            case TokenType.USER:
                auth_key = load_server_info(data_key=WRITE_AUTH_KEY, file=USER_LOCAL_SERVER_FILE)
            case _:
                auth_key = load_server_info(data_key=READONLY_AUTH_KEY, file=USER_LOCAL_SERVER_FILE)
        if auth_key is None:
            return {"error": "Cannot generate selected token mode. Are you the seed node?"}

        watcher_service = load_server_info(data_key=WATCHER_SERVICE_KEY, file=USER_LOCAL_SERVER_FILE)
        public_location = load_server_info(data_key=PUBLIC_LOCATION_KEY, file=USER_LOCAL_SERVER_FILE)

        cluster_token = load_server_info(data_key=CLUSTER_TOKEN_KEY, file=USER_LOCAL_SERVER_FILE) #CLUSTER.get_cluster_token()

        ip_address = load_server_info(SERVER_IP_KEY, file=USER_LOCAL_SERVER_FILE)
        cluster_name = load_server_info(data_key=CLUSTER_NAME_KEY, file=USER_LOCAL_SERVER_FILE)

        join_token = generate_join_token(
            cluster_ip=ip_address,
            cluster_name=cluster_name,
            cluster_token=cluster_token,
            auth_key=auth_key,
            watcher_service=watcher_service,
            public_location=public_location
        )

        return {"token": join_token}
    except Exception as e:
        return {"error": f"Error when generating token: {str(e)}"}

def pool_init(config_values=None):
    """Deploy configured objects to initialise pool"""
    if config_values is None:
        return
    
    # load template config and populate with values
    sidecar_template_yaml = load_template(
        template_path=POOL_CONFIG_TEMPLATE,
        values=config_values)

    try:
        result = request_to_server(
            force_url=FORCE_WATCHER_API_URL,
            force_key=FORCE_WATCHER_API_KEY_URL,
            method="post",
            endpoint="/v1/deploy_generic_model",
            data={"config": sidecar_template_yaml},
            server_creds=USER_LOCAL_SERVER_FILE,
            user_cookie=USER_COOKIE
        )
        return result
    except Exception as e:
        return {"error": f"[red]Error when connecting to kalavai service: {str(e)}"}

def is_connected():
    return is_watcher_alive(server_creds=USER_LOCAL_SERVER_FILE, user_cookie=USER_COOKIE, timeout=10)

def is_agent_running():
    return CLUSTER.is_agent_running()

def is_server():
    return CLUSTER.is_seed_node()

def pause_agent(retries=3):
    try:
        while retries > 0:
            state = CLUSTER.pause_agent()
            if state:
                return {"success"}
            time.sleep(5)
            retries -= 1
    except:
        return {"error": "Could not pause agent"}

def resume_agent(retries=3):
    try:
        while retries > 0:
            state = CLUSTER.restart_agent()
            if state:
                return {"success"}
            time.sleep(5)
            retries -= 1
    except:
        return {"error": "Could not resume agent"}

def cleanup_local():
    safe_remove(CONTAINER_HOST_PATH)
    safe_remove(USER_COMPOSE_FILE)
    safe_remove(USER_VPN_COMPOSE_FILE)
    safe_remove(USER_HELM_APPS_FILE)
    safe_remove(USER_KUBECONFIG_FILE)
    safe_remove(USER_LOCAL_SERVER_FILE)
    safe_remove(USER_TEMPLATES_FOLDER)

def delete_node(name):
    data = {
        "node_names": [name]
    }
    try:
        result = request_to_server(
            force_url=FORCE_WATCHER_API_URL,
            force_key=FORCE_WATCHER_API_KEY_URL,
            method="post",
            endpoint="/v1/delete_nodes",
            data=data,
            server_creds=USER_LOCAL_SERVER_FILE,
            user_cookie=USER_COOKIE
        )
        if result is None or result is True:
            return {f"Node {name} deleted successfully"}
        else:
            return {"error": result}
    except Exception as e:
        return {"error": str(e)}

def stop_pool(skip_node_deletion=False):
    # delete local node from server
    logs = []
    if not skip_node_deletion:
        logs.append(
            delete_node(load_server_info(data_key=NODE_NAME_KEY, file=USER_LOCAL_SERVER_FILE))
        )
    
    # disconnect from VPN first, then remove agent, then remove local files
    try:
        vpns = leave_vpn(container_name=DEFAULT_VPN_CONTAINER_NAME)
        if vpns is not None:
            for vpn in vpns:
                logs.append(f"You have left {vpn} VPN")
    except:
        # no vpn
        pass

    # remove local node agent
    CLUSTER.remove_agent()

    # clean local files
    cleanup_local()

    return {"logs": logs}

def add_node_labels(node_name: str, labels: dict):
    """
    Add labels to a node in the cluster.
    
    Args:
        node_name (str): Name of the node to label
        labels (dict): Dictionary of labels to add to the node
        
    Returns:
        dict: Result of the operation with either success or error message
    """
    data = {
        "node_name": node_name,
        "labels": labels
    }
    try:
        result = request_to_server(
            force_url=FORCE_WATCHER_API_URL,
            force_key=FORCE_WATCHER_API_KEY_URL,
            method="post",
            endpoint="/v1/add_labels_to_node",
            data=data,
            server_creds=USER_LOCAL_SERVER_FILE,
            user_cookie=USER_COOKIE
        )
        if "error" in result:
            return {"error": result["error"]}
        else:
            return {"success": result}

    except Exception as e:
        return {"error": f"Error when adding labels to node {node_name}: {str(e)}"}

def get_node_labels(node_names: list[str]):
    """
    Get labels for specified nodes in the cluster.
    
    Args:
        node_names (list[str]): List of node names to fetch labels from
        
    Returns:
        dict: Result containing the labels for each node or error message
    """
    data = {
        "node_names": node_names
    }
    try:
        result = request_to_server(
            force_url=FORCE_WATCHER_API_URL,
            force_key=FORCE_WATCHER_API_KEY_URL,
            method="post",
            endpoint="/v1/get_node_labels",
            data=data,
            server_creds=USER_LOCAL_SERVER_FILE,
            user_cookie=USER_COOKIE
        )
        if result is not None:
            return {"labels": result}
        else:
            return {"error": "Failed to fetch node labels"}
    except Exception as e:
        return {"error": f"Error when fetching node labels: {str(e)}"}
